from __future__ import annotations

from collections import defaultdict

import numpy as np
import xarray as xr

import tidy3d as td
from tidy3d import Medium
from tidy3d.components.autograd import AutogradFieldMap, get_static
from tidy3d.components.autograd.derivative_utils import DerivativeInfo
from tidy3d.components.data.data_array import DataArray
from tidy3d.config import config
from tidy3d.exceptions import AdjointError
from tidy3d.packaging import disable_local_subpixel

from .utils import E_to_D, get_derivative_maps


def setup_adj(
    data_fields_vjp: AutogradFieldMap,
    sim_data_orig: td.SimulationData,
    sim_fields_keys: list[tuple],
    max_num_adjoint_per_fwd: int,
) -> list[td.Simulation]:
    """Construct an adjoint simulation from a set of data_fields for the VJP."""

    td.log.info("Running custom vjp (adjoint) pipeline.")

    # filter out any data_fields_vjp with exact all 0's
    data_fields_vjp_static = {}
    for k, v in data_fields_vjp.items():
        v_static = get_static(v)
        if np.count_nonzero(v_static) == 0:
            continue
        data_fields_vjp_static[k] = v_static
    data_fields_vjp = data_fields_vjp_static

    for k, v in data_fields_vjp.items():
        if np.any(np.isnan(v)):
            raise AdjointError(
                f"NaN values detected for data field {k} in the adjoint pipeline. This may be "
                f"due to NaN values in the simulation data or the computed value of your "
                f"objective function."
            )

    # if all entries are zero, there is no adjoint sim to run
    if not data_fields_vjp:
        return []

    # start with the full simulation data structure and either zero out the fields
    # that have no tracer data for them or insert the tracer data
    full_sim_data_dict = sim_data_orig._strip_traced_fields(
        include_untraced_data_arrays=True, starting_path=("data",)
    )
    for path in full_sim_data_dict.keys():
        if path in data_fields_vjp:
            full_sim_data_dict[path] = data_fields_vjp[path]
        else:
            full_sim_data_dict[path] *= 0

    # insert the raw VJP data into the .data of the original SimulationData
    sim_data_vjp = sim_data_orig._insert_traced_fields(field_mapping=full_sim_data_dict)

    # make adjoint simulation from that SimulationData
    data_vjp_paths = set(data_fields_vjp.keys())

    num_monitors = len(sim_data_orig.simulation.monitors)
    adjoint_monitors = sim_data_orig.simulation._with_adjoint_monitors(sim_fields_keys).monitors[
        num_monitors:
    ]

    sims_adj = sim_data_vjp._make_adjoint_sims(
        data_vjp_paths=data_vjp_paths,
        adjoint_monitors=adjoint_monitors,
    )

    if len(sims_adj) > max_num_adjoint_per_fwd:
        raise AdjointError(
            f"Number of adjoint simulations ({len(sims_adj)}) exceeds the maximum allowed "
            f"({max_num_adjoint_per_fwd}) per forward simulation. This typically means that "
            "there are many frequencies and monitors in the simulation that are being differentiated "
            "w.r.t. in the objective function. To proceed, please double-check the simulation "
            "setup, increase the 'max_num_adjoint_per_fwd' parameter in the run function, and re-run."
        )

    return sims_adj


def _compute_eps_array(medium: Medium, frequencies: list[float]) -> DataArray:
    """Compute permittivity array for all frequencies."""
    eps_data = [np.mean(medium.eps_model(f)) for f in frequencies]
    return DataArray(data=np.array(eps_data), dims=("f",), coords={"f": frequencies})


def _slice_field_data(
    field_data: dict, freqs: np.ndarray, component_indicator: str | None = None
) -> dict:
    """Slice field data dictionary along frequency dimension."""
    if component_indicator:
        return {k: v.sel(f=freqs) for k, v in field_data.items() if component_indicator in k}
    else:
        return {k: v.sel(f=freqs) for k, v in field_data.items()}


@disable_local_subpixel
def postprocess_adj(
    sim_data_adj: td.SimulationData,
    sim_data_orig: td.SimulationData,
    sim_data_fwd: td.SimulationData,
    sim_fields_keys: list[tuple],
) -> AutogradFieldMap:
    """Postprocess some data from the adjoint simulation into the VJP for the original sim flds."""

    # map of index into 'structures' to the list of paths we need vjps for
    sim_vjp_map = defaultdict(list)
    for _, structure_index, *structure_path in sim_fields_keys:
        structure_path = tuple(structure_path)
        sim_vjp_map[structure_index].append(structure_path)

    # store the derivative values given the forward and adjoint data
    sim_fields_vjp = {}
    for structure_index, structure_paths in sim_vjp_map.items():
        # grab the forward and adjoint data
        fld_fwd = sim_data_fwd._get_adjoint_data(structure_index, data_type="fld")
        eps_fwd = sim_data_fwd._get_adjoint_data(structure_index, data_type="eps")
        fld_adj = sim_data_adj._get_adjoint_data(structure_index, data_type="fld")
        eps_adj = sim_data_adj._get_adjoint_data(structure_index, data_type="eps")

        # post normalize the adjoint fields if a single, broadband source
        fwd_flds_adj_normed = {}
        for key, val in fld_adj.field_components.items():
            fwd_flds_adj_normed[key] = val * sim_data_adj.simulation.post_norm

        fld_adj = fld_adj.updated_copy(**fwd_flds_adj_normed)

        # maps of the E_fwd * E_adj and D_fwd * D_adj, each as as td.FieldData & 'Ex', 'Ey', 'Ez'
        der_maps = get_derivative_maps(
            fld_fwd=fld_fwd,
            eps_fwd=eps_fwd,
            fld_adj=fld_adj,
            eps_adj=eps_adj,
        )
        E_der_map = der_maps["E"]
        D_der_map = der_maps["D"]
        H_der_map = der_maps["H"]

        H_info_exists = H_der_map is not None

        D_fwd = E_to_D(fld_fwd, eps_fwd)
        D_adj = E_to_D(fld_adj, eps_fwd)

        structure = sim_data_fwd.simulation.structures[structure_index]

        # compute epsilon arrays for all frequencies
        # use frequencies from the actual computed derivative map to ensure they exist
        # in both forward and adjoint data (E_der_map = fld_fwd * fld_adj)
        first_field_component = next(iter(E_der_map.field_components.values()))
        adjoint_frequencies = np.array(first_field_component.coords["f"].values)

        monitor_freqs = np.array(fld_adj.monitor.freqs)
        if len(adjoint_frequencies) != len(monitor_freqs) or not np.allclose(
            np.sort(adjoint_frequencies), np.sort(monitor_freqs), rtol=1e-10, atol=0
        ):
            raise ValueError(
                f"Frequency mismatch in adjoint postprocessing for structure {structure_index}. "
                f"Expected frequencies from monitor: {monitor_freqs}, "
                f"but derivative map has: {adjoint_frequencies}. "
            )

        # auto permittivity detection
        sim_orig = sim_data_orig.simulation
        plane_eps = eps_fwd.monitor.geometry
        sim_orig_grid_spec = td.components.grid.grid_spec.GridSpec.from_grid(sim_orig.grid)

        # permittivity without this structure
        structs_no_struct = list(sim_orig.structures)
        structs_no_struct.pop(structure_index)
        sim_no_structure = sim_orig.updated_copy(
            structures=structs_no_struct, monitors=[], sources=[], grid_spec=sim_orig_grid_spec
        )

        # for the outside permittivity of the structure, resize the bounds of the permittivity region
        # to make sure we capture data outside the structure bounds
        low_coords = [center - 0.5 * size for center, size in zip(plane_eps.center, plane_eps.size)]
        high_coords = [
            center + 0.5 * size for center, size in zip(plane_eps.center, plane_eps.size)
        ]

        low_bounds = sim_orig.grid.boundaries.get_bounding_values(low_coords, "left", buffer=1)
        high_bounds = sim_orig.grid.boundaries.get_bounding_values(high_coords, "right", buffer=1)

        resized_center = [0.5 * (low + high) for low, high in zip(low_bounds, high_bounds)]
        resized_size = [(high - low) for low, high in zip(low_bounds, high_bounds)]

        resize_plane_eps = plane_eps.updated_copy(center=resized_center, size=resized_size)

        eps_no_structure_data = [
            sim_no_structure.epsilon(box=resize_plane_eps, coord_key="centers", freq=f)
            for f in adjoint_frequencies
        ]

        eps_no_structure = xr.concat(eps_no_structure_data, dim="f").assign_coords(
            f=adjoint_frequencies
        )

        if structure.medium.is_custom:
            # we can't make an infinite structure from a custom medium permittivity
            eps_inf_structure = None
        else:
            geometry_box = structure.geometry.bounding_box
            background_structures_2d = []
            sim_inf_background_medium = sim_orig.medium
            if np.any(np.array(geometry_box.size) == 0.0):
                zero_coordinate = tuple(geometry_box.size).index(0.0)
                new_size = [td.inf, td.inf, td.inf]
                new_size[zero_coordinate] = 0.0

                background_structures_2d = [
                    structure.updated_copy(geometry=geometry_box.updated_copy(size=new_size))
                ]
            else:
                sim_inf_background_medium = structure.medium

            # permittivity with infinite structure
            structs_inf_struct = list(sim_orig.structures)[structure_index + 1 :]
            sim_inf_structure = sim_orig.updated_copy(
                structures=background_structures_2d + structs_inf_struct,
                medium=sim_inf_background_medium,
                monitors=[],
                sources=[],
                grid_spec=sim_orig_grid_spec,
            )

            eps_inf_structure_data = [
                sim_inf_structure.epsilon(box=plane_eps, coord_key="centers", freq=f)
                for f in adjoint_frequencies
            ]

            eps_inf_structure = xr.concat(eps_inf_structure_data, dim="f").assign_coords(
                f=adjoint_frequencies
            )

        # compute bounds intersection
        struct_bounds = rmin_struct, rmax_struct = structure.geometry.bounds
        rmin_sim, rmax_sim = sim_orig.bounds
        rmin_intersect = tuple([max(a, b) for a, b in zip(rmin_sim, rmin_struct)])
        rmax_intersect = tuple([min(a, b) for a, b in zip(rmax_sim, rmax_struct)])
        bounds_intersect = (rmin_intersect, rmax_intersect)

        # get chunk size - if None, process all frequencies as one chunk
        freq_chunk_size = config.adjoint.solver_freq_chunk_size
        n_freqs = len(adjoint_frequencies)
        if not freq_chunk_size or freq_chunk_size <= 0:
            freq_chunk_size = n_freqs
        else:
            freq_chunk_size = min(freq_chunk_size, n_freqs)

        # process in chunks
        vjp_value_map = {}

        for chunk_start in range(0, n_freqs, freq_chunk_size):
            chunk_end = min(chunk_start + freq_chunk_size, n_freqs)
            freq_slice = slice(chunk_start, chunk_end)

            select_adjoint_freqs = adjoint_frequencies[freq_slice]

            # slice field data for current chunk
            E_der_map_chunk = _slice_field_data(E_der_map.field_components, select_adjoint_freqs)
            D_der_map_chunk = _slice_field_data(D_der_map.field_components, select_adjoint_freqs)
            E_fwd_chunk = _slice_field_data(
                fld_fwd.field_components, select_adjoint_freqs, component_indicator="E"
            )
            E_adj_chunk = _slice_field_data(
                fld_adj.field_components, select_adjoint_freqs, component_indicator="E"
            )
            D_fwd_chunk = _slice_field_data(D_fwd.field_components, select_adjoint_freqs)
            D_adj_chunk = _slice_field_data(D_adj.field_components, select_adjoint_freqs)
            eps_data_chunk = _slice_field_data(eps_fwd.field_components, select_adjoint_freqs)

            H_der_map_chunk = None
            H_fwd_chunk = None
            H_adj_chunk = None

            if H_info_exists:
                H_der_map_chunk = _slice_field_data(
                    H_der_map.field_components, select_adjoint_freqs
                )
                H_fwd_chunk = _slice_field_data(
                    fld_fwd.field_components, select_adjoint_freqs, component_indicator="H"
                )
                H_adj_chunk = _slice_field_data(
                    fld_adj.field_components, select_adjoint_freqs, component_indicator="H"
                )

            # slice epsilon arrays
            eps_no_structure_chunk = (
                eps_no_structure.sel(f=select_adjoint_freqs)
                if eps_no_structure is not None
                else None
            )
            eps_inf_structure_chunk = (
                eps_inf_structure.sel(f=select_adjoint_freqs)
                if eps_inf_structure is not None
                else None
            )

            # create derivative info with sliced data
            derivative_info = DerivativeInfo(
                paths=structure_paths,
                E_der_map=E_der_map_chunk,
                D_der_map=D_der_map_chunk,
                H_der_map=H_der_map_chunk,
                E_fwd=E_fwd_chunk,
                E_adj=E_adj_chunk,
                D_fwd=D_fwd_chunk,
                D_adj=D_adj_chunk,
                H_fwd=H_fwd_chunk,
                H_adj=H_adj_chunk,
                eps_data=eps_data_chunk,
                eps_in=eps_inf_structure_chunk,
                eps_out=eps_no_structure_chunk,
                frequencies=select_adjoint_freqs,  # only chunk frequencies
                bounds=struct_bounds,
                bounds_intersect=bounds_intersect,
                simulation_bounds=sim_data_orig.simulation.bounds,
                is_medium_pec=structure.medium.is_pec,
                background_medium_is_pec=structure.background_medium
                and structure.background_medium.is_pec,
            )

            # compute derivatives for chunk
            vjp_chunk = structure._compute_derivatives(derivative_info)

            # accumulate results
            for path, value in vjp_chunk.items():
                if path in vjp_value_map:
                    val = vjp_value_map[path]
                    if isinstance(val, (list, tuple)) and isinstance(value, (list, tuple)):
                        vjp_value_map[path] = type(val)(x + y for x, y in zip(val, value))
                    else:
                        vjp_value_map[path] += value
                else:
                    vjp_value_map[path] = value

        # store vjps in output map
        for structure_path, vjp_value in vjp_value_map.items():
            sim_path = ("structures", structure_index, *list(structure_path))
            sim_fields_vjp[sim_path] = vjp_value

    return sim_fields_vjp



"""Backwards compatibility - import from tidy3d.components.microwave.path_integrals.integrals instead."""

from __future__ import annotations

from tidy3d.components.microwave.path_integrals.integrals.auto import (
    path_integrals_from_lumped_element,
)

__all__ = [
    "path_integrals_from_lumped_element",
]



"""Typed containers for autograd traced field metadata."""

from __future__ import annotations

import json
from typing import Any, Callable

import pydantic.v1 as pydantic

from tidy3d.components.autograd.types import AutogradFieldMap, dict_ag
from tidy3d.components.base import Tidy3dBaseModel
from tidy3d.components.types import ArrayLike, tidycomplex


class Tracer(Tidy3dBaseModel):
    """Representation of a single traced element within a model."""

    path: tuple[Any, ...] = pydantic.Field(
        ...,
        title="Path to the traced object in the model dictionary.",
    )
    data: float | tidycomplex | ArrayLike = pydantic.Field(..., title="Tracing data")


class FieldMap(Tidy3dBaseModel):
    """Collection of traced elements."""

    tracers: tuple[Tracer, ...] = pydantic.Field(
        ...,
        title="Collection of Tracers.",
    )

    @property
    def to_autograd_field_map(self) -> AutogradFieldMap:
        """Convert to ``AutogradFieldMap`` autograd dictionary."""
        return dict_ag({tracer.path: tracer.data for tracer in self.tracers})

    @classmethod
    def from_autograd_field_map(cls, autograd_field_map: AutogradFieldMap) -> FieldMap:
        """Initialize from an ``AutogradFieldMap`` autograd dictionary."""
        tracers = []
        for path, data in autograd_field_map.items():
            tracers.append(Tracer(path=path, data=data))
        return cls(tracers=tuple(tracers))


def _encoded_path(path: tuple[Any, ...]) -> str:
    """Return a stable JSON representation for a traced path."""
    return json.dumps(list(path), separators=(",", ":"), ensure_ascii=True)


class TracerKeys(Tidy3dBaseModel):
    """Collection of traced field paths."""

    keys: tuple[tuple[Any, ...], ...] = pydantic.Field(
        ...,
        title="Collection of tracer keys.",
    )

    def encoded_keys(self) -> list[str]:
        """Return the JSON-encoded representation of keys."""
        return [_encoded_path(path) for path in self.keys]

    @classmethod
    def from_field_mapping(
        cls,
        field_mapping: AutogradFieldMap,
        *,
        sort_key: Callable[[tuple[Any, ...]], str] | None = None,
    ) -> TracerKeys:
        """Construct keys from an autograd field mapping."""
        if sort_key is None:
            sort_key = _encoded_path

        sorted_paths = tuple(sorted(field_mapping.keys(), key=sort_key))
        return cls(keys=sorted_paths)



"""Specification for modes associated with transmission lines."""

from __future__ import annotations

from typing import Optional, Union

import numpy as np
import pydantic.v1 as pd

from tidy3d.components.base import cached_property
from tidy3d.components.geometry.base import Box
from tidy3d.components.geometry.bound_ops import bounds_contains
from tidy3d.components.microwave.base import MicrowaveBaseModel
from tidy3d.components.microwave.path_integrals.specs.impedance import (
    AutoImpedanceSpec,
    ImpedanceSpecType,
)
from tidy3d.components.mode_spec import AbstractModeSpec
from tidy3d.components.types import annotate_type
from tidy3d.constants import fp_eps
from tidy3d.exceptions import SetupError

TEM_POLARIZATION_THRESHOLD = 0.995
QTEM_POLARIZATION_THRESHOLD = 0.95


class MicrowaveModeSpec(AbstractModeSpec, MicrowaveBaseModel):
    """Specification for transmission line modes and microwave waveguides.

    Notes
    -----
        The :class:`.MicrowaveModeSpec` class specifies how quantities related to transmission line
        modes and microwave waveguides are computed. For example, it defines the paths for line integrals, which are used to
        compute voltage, current, and characteristic impedance of the transmission line.

    Example
    -------
    >>> import tidy3d as td
    >>> # Using automatic impedance calculation (single spec, will be duplicated for all modes)
    >>> mode_spec_auto = td.MicrowaveModeSpec(
    ...     num_modes=2,
    ...     impedance_specs=td.AutoImpedanceSpec()
    ... )
    >>> # Using custom impedance specification for multiple modes
    >>> voltage_spec = td.AxisAlignedVoltageIntegralSpec(
    ...     center=(0, 0, 0), size=(0, 0, 1), sign="+"
    ... )
    >>> current_spec = td.AxisAlignedCurrentIntegralSpec(
    ...     center=(0, 0, 0), size=(2, 1, 0), sign="+"
    ... )
    >>> custom_impedance = td.CustomImpedanceSpec(
    ...     voltage_spec=voltage_spec, current_spec=current_spec
    ... )
    >>> mode_spec_custom = td.MicrowaveModeSpec(
    ...     num_modes=1,
    ...     impedance_specs=custom_impedance
    ... )
    """

    impedance_specs: Union[
        annotate_type(ImpedanceSpecType),
        tuple[Optional[annotate_type(ImpedanceSpecType)], ...],
    ] = pd.Field(
        default_factory=AutoImpedanceSpec._default_without_license_warning,
        title="Impedance Specifications",
        description="Field controls how the impedance is calculated for each mode calculated by the mode solver. "
        "Can be a single impedance specification (which will be applied to all modes) or a tuple of specifications "
        "(one per mode). The number of impedance specifications should match the number of modes field. "
        "When an impedance specification of ``None`` is used, the impedance calculation will be "
        "ignored for the associated mode.",
    )

    tem_polarization_threshold: float = pd.Field(
        TEM_POLARIZATION_THRESHOLD,
        gt=0.0,
        le=1.0,
        title="TEM Polarization Threshold",
        description="Threshold for classifying modes as TEM, TE, or TM based on mean TE/TM fraction "
        "across frequencies. A mode is classified as TEM if both mean TE and TM fractions are greater "
        "than or equal to this threshold. Similarly, a mode is classified as TE (or TM) if the mean TE "
        "(or TM) fraction is greater than or equal to this threshold.",
    )

    qtem_polarization_threshold: float = pd.Field(
        QTEM_POLARIZATION_THRESHOLD,
        gt=0.0,
        le=1.0,
        title="Quasi-TEM Polarization Threshold",
        description="Threshold for classifying modes as quasi-TEM based on TE/TM fraction at the lowest "
        "frequency. A mode is classified as quasi-TEM if both TE and TM fractions at the lowest frequency "
        "are greater than or equal to this threshold.",
    )

    @cached_property
    def _impedance_specs_as_tuple(self) -> tuple[Optional[ImpedanceSpecType]]:
        """Gets the impedance_specs field converted to a tuple."""
        if isinstance(self.impedance_specs, Union[tuple, list]):
            return tuple(self.impedance_specs)
        return (self.impedance_specs,)

    @cached_property
    def _using_auto_current_spec(self) -> bool:
        """Checks whether at least one of the modes will require an auto setup of the current path specification."""
        return any(
            isinstance(impedance_spec, AutoImpedanceSpec)
            for impedance_spec in self._impedance_specs_as_tuple
        )

    @pd.validator("impedance_specs", always=True)
    def check_impedance_specs_consistent_with_num_modes(cls, val, values):
        """Check that the number of impedance specifications is equal to the number of modes.
        A single impedance spec is also permitted."""
        num_modes = values.get("num_modes")
        if isinstance(val, Union[tuple, list]):
            num_impedance_specs = len(val)
        else:
            return val

        # Otherwise, check that the count matches
        if num_impedance_specs != num_modes:
            raise SetupError(
                f"Given {num_impedance_specs} impedance specifications in the 'MicrowaveModeSpec', "
                f"but the number of modes requested is {num_modes}. Please ensure that the "
                "number of impedance specifications is equal to the number of modes, or provide "
                "a single specification to apply to all modes."
            )

        return val

    def _check_path_integrals_within_box(self, box: Box):
        """Raise SetupError if a ``CustomImpedanceSpec`` includes a path specification
        defined outside a candidate box.
        """
        for impedance_ind, impedance_spec in enumerate(self._impedance_specs_as_tuple):
            if isinstance(impedance_spec, AutoImpedanceSpec) or impedance_spec is None:
                continue

            # Check both voltage and current specs using the same logic
            specs_to_check = [
                (impedance_spec.voltage_spec, "voltage"),
                (impedance_spec.current_spec, "current"),
            ]

            for spec, spec_type in specs_to_check:
                if spec is None:
                    continue

                box_bounds = box.bounds
                # If the box is a plane (one dimension is zero), we need to ignore
                # the bounds check along the normal axis
                if box.size.count(0.0) == 1:
                    normal_axis = box._normal_axis
                    # Convert tuple to list so we can modify it
                    box_bounds = [list(box_bounds[0]), list(box_bounds[1])]
                    # Set the bounds along normal axis to match the spec bounds
                    box_bounds[0][normal_axis] = spec.bounds[0][normal_axis]
                    box_bounds[1][normal_axis] = spec.bounds[1][normal_axis]
                    # Convert back to tuple for bounds_contains
                    box_bounds = (tuple(box_bounds[0]), tuple(box_bounds[1]))

                if not bounds_contains(
                    box_bounds, spec.bounds, fp_eps, np.finfo(np.float32).smallest_normal
                ):
                    raise SetupError(
                        "A 'MicrowaveModeSpec' must be setup with all path specifications defined within "
                        f"the bounds of the mode solving plane. The 'CustomImpedanceSpec' at index "
                        f"'{impedance_ind}' was provided with a {spec_type} path specification with bounds "
                        f"'{spec.bounds}', but the mode plane bounds are '{box.bounds}'."
                    )



"""Class for representing a scattering matrix wave port."""

from __future__ import annotations

from typing import Optional, Union

import pydantic.v1 as pd

from tidy3d.components.base import cached_property, skip_if_fields_missing
from tidy3d.components.boundary import ABCBoundary, InternalAbsorber, ModeABCBoundary
from tidy3d.components.data.data_array import FreqDataArray, FreqModeDataArray
from tidy3d.components.data.sim_data import SimulationData
from tidy3d.components.geometry.base import Box
from tidy3d.components.grid.grid import Grid
from tidy3d.components.microwave.data.monitor_data import MicrowaveModeData
from tidy3d.components.microwave.mode_spec import MicrowaveModeSpec
from tidy3d.components.microwave.monitor import MicrowaveModeMonitor
from tidy3d.components.simulation import Simulation
from tidy3d.components.source.field import ModeSource
from tidy3d.components.source.frame import PECFrame
from tidy3d.components.source.time import GaussianPulse
from tidy3d.components.structure import MeshOverrideStructure
from tidy3d.components.types import Axis, Direction, FreqArray
from tidy3d.exceptions import SetupError, ValidationError
from tidy3d.log import log
from tidy3d.plugins.mode import ModeSolver
from tidy3d.plugins.smatrix.ports.base_terminal import AbstractTerminalPort

DEFAULT_WAVE_PORT_NUM_CELLS = 5
MIN_WAVE_PORT_NUM_CELLS = 3
DEFAULT_WAVE_PORT_FRAME = PECFrame()


class WavePort(AbstractTerminalPort, Box):
    """Class representing a single wave port"""

    direction: Direction = pd.Field(
        ...,
        title="Direction",
        description="'+' or '-', defining which direction is considered 'input'.",
    )

    mode_spec: MicrowaveModeSpec = pd.Field(
        default_factory=MicrowaveModeSpec._default_without_license_warning,
        title="Mode Specification",
        description="Parameters to feed to mode solver which determine modes and how transmission line "
        "quantities, e.g., charateristic impedance, are computed.",
    )

    num_grid_cells: Optional[int] = pd.Field(
        DEFAULT_WAVE_PORT_NUM_CELLS,
        ge=MIN_WAVE_PORT_NUM_CELLS,
        title="Number of Grid Cells",
        description="Number of mesh grid cells in the transverse plane of the `WavePort`. "
        "Used in generating the suggested list of :class:`.MeshOverrideStructure` objects. "
        "Must be greater than or equal to 3. When set to `None`, no grid refinement is performed.",
    )

    conjugated_dot_product: bool = pd.Field(
        False,
        title="Conjugated Dot Product",
        description="Use conjugated or non-conjugated dot product for mode decomposition.",
    )

    frame: Optional[PECFrame] = pd.Field(
        DEFAULT_WAVE_PORT_FRAME,
        title="Source Frame",
        description="Add a thin frame around the source during FDTD run for an improved injection.",
    )

    absorber: Union[bool, ABCBoundary, ModeABCBoundary] = pd.Field(
        True,
        title="Absorber",
        description="Place a mode absorber in the port. If ``True``, an automatically generated mode absorber is placed in the port. "
        "If :class:`.ABCBoundary` or :class:`.ModeABCBoundary`, a mode absorber is placed in the port with the specified boundary conditions.",
    )

    extrude_structures: bool = pd.Field(
        False,
        title="Extrude Structures",
        description="Extrudes structures that intersect the wave port plane by a few grid cells when ``True``, improving mode injection accuracy.",
    )

    mode_index: Optional[pd.NonNegativeInt] = pd.Field(
        None,
        title="Mode Index (deprecated)",
        description="Index into the collection of modes returned by mode solver. "
        "Specifies which mode to inject using this port. "
        "Deprecated. Use the 'mode_selection' field instead.",
    )

    mode_selection: Optional[tuple[int, ...]] = pd.Field(
        None,
        title="Mode Selection",
        description="Selects specific mode(s) to use from the mode solver. "
        "Can be a single integer for one mode, or a tuple of integers for multiple modes. "
        "If ``None`` (default), all modes from the ``mode_spec`` are used. "
        "Indices must be non-negative and less than ``mode_spec.num_modes``.",
    )

    @cached_property
    def injection_axis(self) -> Axis:
        """Injection axis of the port."""
        return self.size.index(0.0)

    @cached_property
    def transverse_axes(self) -> tuple[Axis, Axis]:
        """Transverse axes of the port."""
        _, trans_axes = Box.pop_axis([0, 1, 2], self.injection_axis)
        return trans_axes

    @cached_property
    def _mode_monitor_name(self) -> str:
        """Return the name of the :class:`.MicrowaveModeMonitor` associated with this port."""
        return f"{self.name}_mode"

    @cached_property
    def _mode_indices(self) -> tuple[int, ...]:
        """Mode indices that will be excited/monitored by this port."""
        if self.mode_index is not None and self.mode_selection is None:
            return (self.mode_index,)
        if self.mode_selection is not None:
            # User specified specific modes
            return self.mode_selection
        # Default: use all modes
        return tuple(range(self.mode_spec.num_modes))

    def to_source(
        self,
        source_time: GaussianPulse,
        snap_center: Optional[float] = None,
        mode_index: int = 0,
    ) -> ModeSource:
        """Create a mode source from the wave port."""
        center = list(self.center)
        if snap_center:
            center[self.injection_axis] = snap_center
        return ModeSource(
            center=center,
            size=self.size,
            source_time=source_time,
            mode_spec=self.mode_spec,
            mode_index=mode_index,
            direction=self.direction,
            name=self.name,
            frame=self.frame,
        )

    def to_monitors(
        self, freqs: FreqArray, snap_center: Optional[float] = None, grid: Grid = None
    ) -> list[MicrowaveModeMonitor]:
        """The wave port uses a :class:`.MicrowaveModeMonitor` to compute the characteristic impedance
        and the port voltages and currents."""
        center = list(self.center)
        if snap_center:
            center[self.injection_axis] = snap_center
        mode_mon = MicrowaveModeMonitor(
            center=self.center,
            size=self.size,
            freqs=freqs,
            name=self._mode_monitor_name,
            colocate=False,
            mode_spec=self.mode_spec,
            store_fields_direction=self.direction,
            conjugated_dot_product=self.conjugated_dot_product,
        )
        return [mode_mon]

    def to_mode_solver(self, simulation: Simulation, freqs: FreqArray) -> ModeSolver:
        """Helper to create a :class:`.ModeSolver` instance."""
        mode_solver = ModeSolver(
            simulation=simulation,
            plane=self.geometry,
            mode_spec=self.mode_spec,
            freqs=freqs,
            direction=self.direction,
            colocate=False,
        )
        return mode_solver

    def to_absorber(
        self, snap_center: Optional[float] = None, freq_spec: Optional[pd.NonNegativeFloat] = None
    ) -> InternalAbsorber:
        """Create an internal absorber from the wave port."""
        center = list(self.center)
        if snap_center:
            center[self.injection_axis] = snap_center
        if isinstance(self.absorber, (ABCBoundary, ModeABCBoundary)):
            boundary_spec = self.absorber
        else:
            # TODO: ModeABCBoundary currently only accepts one mode, so
            # we choose the first mode for now
            mode_index = self._mode_indices[0]
            boundary_spec = ModeABCBoundary(
                mode_spec=self.mode_spec,
                mode_index=mode_index,
                plane=self.geometry,
                freq_spec=freq_spec,
            )
        return InternalAbsorber(
            center=center,
            size=self.size,
            boundary_spec=boundary_spec,
            direction="-"
            if self.direction == "+"
            else "+",  # absorb in the opposite direction of source
            grid_shift=1,  # absorb in the next pixel
        )

    def compute_voltage(self, sim_data: SimulationData) -> FreqDataArray:
        """Helper to compute voltage across the port."""
        mode_data: MicrowaveModeData = sim_data[self._mode_monitor_name]
        voltage_coeffs = mode_data.transmission_line_data.voltage_coeffs
        amps = mode_data.amps
        fwd_amps = amps.sel(direction="+").squeeze()
        bwd_amps = amps.sel(direction="-").squeeze()
        return voltage_coeffs * (fwd_amps + bwd_amps)

    def compute_current(self, sim_data: SimulationData) -> FreqDataArray:
        """Helper to compute current flowing through the port."""
        mode_data: MicrowaveModeData = sim_data[self._mode_monitor_name]
        current_coeffs = mode_data.transmission_line_data.current_coeffs
        amps = mode_data.amps
        fwd_amps = amps.sel(direction="+").squeeze()
        bwd_amps = amps.sel(direction="-").squeeze()
        # In ModeData, fwd_amps and bwd_amps are not relative to
        # the direction fields are stored
        sign = 1.0
        if self.direction == "-":
            sign = -1.0
        return sign * current_coeffs * (fwd_amps - bwd_amps)

    def get_port_impedance(
        self, sim_mode_data: Union[SimulationData, MicrowaveModeData], mode_index: int
    ) -> FreqModeDataArray:
        """Retrieve the characteristic impedance of the port for a specific mode.

        The port impedance is computed from the transmission line mode characteristics,
        which should ideally be TEM (Transverse Electromagnetic) or at least quasi-TEM.
        The impedance is extracted from the transmission line data computed by the
        mode solver.

        Parameters
        ----------
        sim_mode_data : Union[:class:`.SimulationData`, :class:`.MicrowaveModeData`]
            Simulation data containing the mode monitor results, or the mode data directly.
            If :class:`.SimulationData` is provided, the mode data is extracted using the
            port's mode monitor name.
        mode_index : int
            Index of the mode for which to compute the impedance. This selects a specific
            mode from the mode spectrum computed by the mode solver.

        Returns
        -------
        :class:`.FreqModeDataArray`
            Frequency-dependent characteristic impedance Z0 for the specified mode.
            The impedance is complex-valued and varies with frequency.
        """
        if isinstance(sim_mode_data, SimulationData):
            mode_data = sim_mode_data[self._mode_monitor_name]
        else:
            mode_data = sim_mode_data
        return mode_data.transmission_line_data.Z0.sel(mode_index=[mode_index])

    def to_mesh_overrides(self) -> list[MeshOverrideStructure]:
        """Creates a list of :class:`.MeshOverrideStructure` for mesh refinement in the transverse
        plane of the port. The mode source requires at least 3 grid cells in the transverse
        dimensions, so these mesh overrides will be added to the simulation to ensure that this
        requirement is satisfied.
        """
        dl = [None] * 3
        for trans_axis in self.transverse_axes:
            dl[trans_axis] = self.size[trans_axis] / self.num_grid_cells

        return [
            MeshOverrideStructure(
                geometry=Box(center=self.center, size=self.size),
                dl=dl,
                shadow=False,
                priority=-1,
            )
        ]

    @pd.validator("mode_spec", always=True)
    def _validate_path_integrals_within_port(cls, val, values):
        """Validate that the microwave mode spec contains path specs all within the port bounds."""
        center = values["center"]
        size = values["size"]
        self_plane = Box(size=size, center=center)
        try:
            val._check_path_integrals_within_box(self_plane)
        except SetupError as e:
            raise SetupError(
                f"Failed to setup '{cls.__name__}' with the suppled 'MicrowaveModeSpec'. {e!s}"
            ) from e
        return val

    @pd.validator("mode_selection", always=True)
    @skip_if_fields_missing(["mode_spec"])
    def _validate_mode_selection(cls, val, values):
        """Validate that mode_selection contains valid, unique indices within range."""
        if val is None:
            return val

        indices = val

        # Check for non-negative integers
        if any(idx < 0 for idx in indices):
            raise ValidationError(
                f"'mode_selection' must contain non-negative integers. Got: {indices}"
            )

        # Check for duplicates
        if len(indices) != len(set(indices)):
            duplicates = [idx for idx in set(indices) if list(indices).count(idx) > 1]
            raise ValidationError(
                f"'mode_selection' contains duplicate entries: {duplicates}. "
                "Each index must appear only once."
            )

        # Check that indices are within range of num_modes
        mode_spec = values["mode_spec"]
        num_modes = mode_spec.num_modes
        invalid_indices = [idx for idx in indices if idx >= num_modes]
        if invalid_indices:
            raise ValidationError(
                f"'mode_selection' contains indices {invalid_indices} that are >= "
                f"'mode_spec.num_modes' ({num_modes}). Valid range is 0 to {num_modes - 1}."
            )

        return val

    @pd.root_validator(pre=False)
    def _check_absorber_if_extruding_structures(cls, values):
        """Raise validation error when ``extrude_structures`` is set to ``True``
        while ``absorber`` is set to ``False``."""

        if values.get("extrude_structures") and not values.get("absorber"):
            raise ValidationError(
                "Structure extrusion for a waveport requires an internal absorber. Set `absorber=True` to enable it."
            )

        return values

    @pd.validator("mode_index", always=True)
    def _mode_index_deprecated(cls, val):
        """Warn that 'mode_index' is deprecated in favor of 'mode_selection'."""
        if val is not None:
            log.warning(
                "'mode_index' is deprecated and will be removed in future versions. "
                "Please use 'mode_selection' instead."
            )
        return val

    @pd.validator("mode_index", always=True)
    @skip_if_fields_missing(["mode_spec"])
    def _validate_mode_index(cls, val, values):
        """Validate that mode_selection contains valid, unique indices within range."""
        if val is None:
            return val
        num_modes = values["mode_spec"].num_modes
        if val >= num_modes:
            raise ValidationError(
                f"'mode_index' is >= "
                f"'mode_spec.num_modes' ({num_modes}). Valid range is 0 to {num_modes - 1}."
            )
        return val

    @property
    def _is_using_mesh_refinement(self) -> bool:
        """Check if this wave port is using mesh refinement options.

        Returns ``True`` if a custom grid cell count is specified.
        """
        return self.num_grid_cells is not None



"""Mode solver for propagating EM modes."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

import numpy as np

from tidy3d.components.base import Tidy3dBaseModel
from tidy3d.components.types import EpsSpecType, ModeSolverType, Numpy
from tidy3d.constants import C_0, ETA_0, fp_eps, pec_val

from .derivatives import create_d_matrices as d_mats
from .derivatives import create_s_matrices as s_mats
from .transforms import angled_transform, radial_transform

# Consider vec to be complex if norm(vec.imag)/norm(vec) > TOL_COMPLEX
TOL_COMPLEX = 1e-10
# Tolerance for eigs
TOL_EIGS = fp_eps / 10
# Relative tolerance of complex effective index for finding candidate degenerate modes
TOL_DEGENERATE_CANDIDATE = 1e3 * TOL_EIGS
# Tolerance used to determine whether possible degenerate modes need manual orthogonalization
TOL_NEEDS_ORTHOGONALIZATION = 1e2 * TOL_EIGS
# Tolerance for deciding on the matrix to be diagonal or tensorial
TOL_TENSORIAL = 1e-6
# shift target neff by this value, both rel and abs, whichever results in larger shift
TARGET_SHIFT = 10 * fp_eps
# Preconditioner: "Jacobi" or "Material" based
PRECONDITIONER = "Material"
# Good conductor permittivity cut-off value. Let it be as large as possible so long as not causing overflow in
# double precision. This value is very heuristic.
GOOD_CONDUCTOR_CUT_OFF = 1e70

if TYPE_CHECKING:
    from scipy import sparse as sp

# Consider a material to be good conductor if |ep| (or |mu|) > GOOD_CONDUCTOR_THRESHOLD * |pec_val|
GOOD_CONDUCTOR_THRESHOLD = 0.9


class EigSolver(Tidy3dBaseModel):
    """Interface for computing eigenvalues given permittivity and mode spec.
    It's a collection of static methods.
    """

    @classmethod
    def compute_modes(
        cls,
        eps_cross,
        coords,
        freq,
        mode_spec,
        precision,
        mu_cross=None,
        split_curl_scaling=None,
        symmetry=(0, 0),
        direction="+",
        solver_basis_fields=None,
        plane_center: Optional[tuple[float, float]] = None,
    ) -> tuple[Numpy, Numpy, EpsSpecType]:
        """
        Solve for the modes of a waveguide cross-section.

        Parameters
        ----------
        eps_cross : array_like or tuple of array_like
            Either a single 2D array defining the relative permittivity in the cross-section,
            or nine 2D arrays defining the permittivity at the Ex, Ey, and Ez locations
            of the Yee grid in the order xx, xy, xz, yx, yy, yz, zx, zy, zz.
        coords : List[Numpy]
            Two 1D arrays with each with size one larger than the corresponding axis of
            ``eps_cross``.
            Defines a (potentially non-uniform) Cartesian grid on which the modes are computed.
        freq : float
            (Hertz) Frequency at which the eigenmodes are computed.
        mode_spec : ModeSpec
            ``ModeSpec`` object containing specifications of the mode solver.
        precision : Literal["single", "double"]
            Apply "single" or "double" precision in mode solver.
        mu_cross : array_like or tuple of array_like
            Either a single 2D array defining the relative permeability in the cross-section,
            or nine 2D arrays defining the permeability at the Hx, Hy, and Hz locations
            of the Yee grid in the order xx, xy, xz, yx, yy, yz, zx, zy, zz.
            Set to 1 if `None`.
        split_curl_scaling : tuple of array_like
            Split curl coefficient to Curl E. Three 2D arrays defining the scaling factor
            at the Ex, Ey, and Ez locations of the Yee grid in the order xx, yy, zz.
            Following Benkler's approach, we formulate it as the following:
            1) during mode solver: eps_cross -> eps_corss / scaling, so eigenvector is E * scaling
            2) in postprocessing: apply scaling^-1 to eigenvector to obtain E
        direction : Union["+", "-"]
            Direction of mode propagation.
        solver_basis_fields
            If provided, solve for modes in this basis.
        plane_center
            The center of the mode plane along the tangential axes of the global simulation. Used
            in case of bend modes to offset the coordinates correctly w.r.t. the bend radius, which
            is assumed to refer to the distance from the bend center to the mode plane center.

        Returns
        -------
        Tuple[Numpy, Numpy, str]
            The first array gives the E and H fields for all modes, the second one gives the complex
            effective index. The last variable describes permittivity characterization on the mode
            solver's plane ("diagonal", "tensorial_real", or "tensorial_complex").
        """

        num_modes = mode_spec.num_modes
        bend_radius = mode_spec.bend_radius
        bend_axis = mode_spec.bend_axis
        angle_theta = mode_spec.angle_theta
        angle_phi = mode_spec.angle_phi
        omega = 2 * np.pi * freq
        k0 = omega / C_0
        enable_incidence_matrices = False  # Experimental feature, always off for now

        eps_formated = cls.format_medium_data(eps_cross)
        eps_xx, eps_xy, eps_xz, eps_yx, eps_yy, eps_yz, eps_zx, eps_zy, eps_zz = eps_formated

        mu_formated = None
        if mu_cross is not None:
            mu_formated = cls.format_medium_data(mu_cross)

        Nxy = eps_xx.shape
        Nx, Ny = Nxy
        N = eps_xx.size

        if len(coords[0]) != Nx + 1 or len(coords[1]) != Ny + 1:
            raise ValueError("Mismatch between 'coords' and 'esp_cross' shapes.")
        new_coords = [np.copy(c) for c in coords]

        """We work with full tensorial epsilon in mu to handle the most general cases that can
        be introduced by coordinate transformations. In the solver, we distinguish the case when
        these tensors are still diagonal, in which case the matrix for diagonalization has shape
        (2N, 2N), and the full tensorial case, in which case it has shape (4N, 4N)."""
        eps_tensor = np.zeros((3, 3, N), dtype=np.complex128)
        mu_tensor = np.zeros((3, 3, N), dtype=np.complex128)
        identity_tensor = np.zeros((3, 3, N), dtype=np.complex128)
        for row, eps_row in enumerate(
            [[eps_xx, eps_xy, eps_xz], [eps_yx, eps_yy, eps_yz], [eps_zx, eps_zy, eps_zz]]
        ):
            identity_tensor[row, row, :] = 1.0
            for col, eps in enumerate(eps_row):
                if split_curl_scaling is not None and col == row:
                    outside_pec = ~np.isclose(split_curl_scaling[col], 0)
                    eps[outside_pec] /= split_curl_scaling[col][outside_pec]

                eps_tensor[row, col, :] = eps.ravel()

        if mu_formated is not None:
            mu_xx, mu_xy, mu_xz, mu_yx, mu_yy, mu_yz, mu_zx, mu_zy, mu_zz = mu_formated
            for row, mu_row in enumerate(
                [[mu_xx, mu_xy, mu_xz], [mu_yx, mu_yy, mu_yz], [mu_zx, mu_zy, mu_zz]]
            ):
                for col, mu in enumerate(mu_row):
                    mu_tensor[row, col, :] = mu.ravel()
        else:
            mu_tensor = np.copy(identity_tensor)

        # Get Jacobian of all coordinate transformations. Initialize as identity (same as mu so far)
        jac_e = np.real(np.copy(identity_tensor))
        jac_h = np.real(np.copy(identity_tensor))

        if np.abs(angle_theta) > 0:
            new_coords, jac_e, jac_h = angled_transform(new_coords, angle_theta, angle_phi)

        if bend_radius is not None:
            if plane_center is None:
                raise ValueError(
                    "When using a nonzero 'bend_radius', the 'plane_center' must also "
                    "be provided to correctly offset the coordinates."
                )
            new_coords, jac_e_tmp, jac_h_tmp = radial_transform(
                new_coords, bend_radius, bend_axis, plane_center
            )
            jac_e = np.einsum("ij...,jp...->ip...", jac_e_tmp, jac_e)
            jac_h = np.einsum("ij...,jp...->ip...", jac_h_tmp, jac_h)

        """We also need to keep track of the transformation of the k-vector. This is
        the eigenvalue of the momentum operator assuming some sort of translational invariance and is
        different from just the transformation of the derivative operator. For example, in a bent
        waveguide, there is strictly speaking no k-vector in the original coordinates as the system
        is not translationally invariant there. However, if we define kz = R k_phi, then the
        effective index approaches that for a straight-waveguide in the limit of infinite radius.
        Since we use w = R phi in the radial_transform, there is nothing else needed in the k transform.
        For the angled_transform, the transformation between k-vectors follows from writing the field as
        E' exp(i k_p w) in transformed coordinates, and identifying this with
        E exp(i k_x x + i k_y y + i k_z z) in the original ones."""
        kxy = np.cos(angle_theta) ** 2
        kz = np.cos(angle_theta) * np.sin(angle_theta)
        kp_to_k = np.array([kxy * np.sin(angle_phi), kxy * np.cos(angle_phi), kz])

        # Transform epsilon and mu
        jac_e_det = np.linalg.det(np.moveaxis(jac_e, [0, 1], [-2, -1]))
        jac_h_det = np.linalg.det(np.moveaxis(jac_h, [0, 1], [-2, -1]))
        eps_tensor = np.einsum("ij...,jp...->ip...", jac_e, eps_tensor)  # J.dot(eps)
        eps_tensor = np.einsum("ij...,pj...->ip...", eps_tensor, jac_e)  # (J.dot(eps)).dot(J.T)
        eps_tensor /= jac_e_det
        mu_tensor = np.einsum("ij...,jp...->ip...", jac_h, mu_tensor)
        mu_tensor = np.einsum("ij...,pj...->ip...", mu_tensor, jac_h)
        mu_tensor /= jac_h_det

        # # Uncomment block to force eps and mu to be translationally invariant into the PML.
        # # This may be important for bends as the jacobian transformation breaks the invariance, but
        # # tests reveal that it has little effect.
        # eps_tensor = cls.make_pml_invariant(Nxy, eps_tensor, mode_spec.num_pml)
        # mu_tensor = cls.make_pml_invariant(Nxy, mu_tensor, mode_spec.num_pml)

        """ Boundaries are imposed through the derivative matrices. The forward derivative matrices
        always impose PEC boundary at the xmax and ymax interfaces, and on the xmin and ymin
        interfaces unless PMC symmetry is present. If so, the PMC boundary is imposed through the
        backward derivative matrices."""
        dmin_pmc = [sym == 1 for sym in symmetry]

        # Primal grid steps for E-field derivatives
        dl_f = [new_cs[1:] - new_cs[:-1] for new_cs in new_coords]
        # Dual grid steps for H-field derivatives
        dl_tmp = [(dl[:-1] + dl[1:]) / 2 for dl in dl_f]
        dl_b = [np.hstack((d1[0], d2)) for d1, d2 in zip(dl_f, dl_tmp)]
        dls = (dl_f, dl_b)

        # Derivative matrices with PEC boundaries by default and optional PMC at the near end
        der_mats_tmp = d_mats(Nxy, dls, dmin_pmc)

        # PML matrices; do not impose PML on the bottom when symmetry present
        dmin_pml = np.array(symmetry) == 0
        pml_mats = s_mats(omega, Nxy, mode_spec.num_pml, dls, eps_tensor, mu_tensor, dmin_pml)

        # Add the PML on top of the derivatives; normalize by k0 to match the EM-possible notation
        der_mats = [Smat.dot(Dmat) / k0 for Smat, Dmat in zip(pml_mats, der_mats_tmp)]

        # Determine initial guess value for the solver in transformed coordinates
        if mode_spec.target_neff is None:
            eps_physical = np.array(eps_cross)
            eps_physical = eps_physical[np.abs(eps_physical) < np.abs(pec_val)]
            n_max = np.sqrt(np.max(np.abs(eps_physical)))
            target = n_max
        else:
            target = mode_spec.target_neff
        target_neff_p = target / np.linalg.norm(kp_to_k)

        # shift target_neff slightly to avoid cases where the shiftted matrix is exactly singular
        if abs(TARGET_SHIFT) > abs(target_neff_p * TARGET_SHIFT):
            target_neff_p += TARGET_SHIFT
        else:
            target_neff_p *= 1 + TARGET_SHIFT

        # preprocess solver_basis_fields
        basis_E = None
        if solver_basis_fields is not None:
            basis_E = solver_basis_fields[:3, ...]
            try:
                basis_E = basis_E.reshape((3, Nx * Ny, num_modes))
            except ValueError:
                raise ValueError(
                    "Shape mismatch between 'basis_fields' and requested mode data. "
                    "Make sure the mode solvers are set up the same, and that the "
                    "basis mode solver data has 'colocate=False'."
                ) from None
            if split_curl_scaling is not None:
                basis_E = cls.split_curl_field_postprocess_inverse(split_curl_scaling, basis_E)
            jac_e_inv = np.moveaxis(
                np.linalg.inv(np.moveaxis(jac_e, [0, 1], [-2, -1])), [-2, -1], [0, 1]
            )
            basis_E = np.sum(jac_e_inv[..., None] * basis_E[:, None, ...], axis=0)

        # Solve for the modes
        E, H, neff, keff, eps_spec = cls.solver_em(
            Nx,
            Ny,
            eps_tensor,
            mu_tensor,
            der_mats,
            num_modes,
            target_neff_p,
            precision,
            direction,
            enable_incidence_matrices,
            basis_E=basis_E,
            dls=dls,
            dmin_pmc=dmin_pmc,
        )

        # Transform back to original axes, E = J^T E'
        E = np.sum(jac_e[..., None] * E[:, None, ...], axis=0)
        if split_curl_scaling is not None:
            E = cls.split_curl_field_postprocess(split_curl_scaling, E)
        E = E.reshape((3, Nx, Ny, 1, num_modes))
        H = np.sum(jac_h[..., None] * H[:, None, ...], axis=0)
        H = H.reshape((3, Nx, Ny, 1, num_modes))

        fields = np.stack((E, H), axis=0)

        neff = neff * np.linalg.norm(kp_to_k)
        keff = keff * np.linalg.norm(kp_to_k)

        if mode_spec.precision == "single":
            # Recast to single precision which may have changed due to earlier manipulations
            fields = fields.astype(np.complex64)

        return fields, neff + 1j * keff, eps_spec

    @classmethod
    def solver_em(
        cls,
        Nx,
        Ny,
        eps_tensor,
        mu_tensor,
        der_mats,
        num_modes,
        neff_guess,
        mat_precision,
        direction,
        enable_incidence_matrices,
        basis_E,
        dls,
        dmin_pmc=None,
    ):
        """Solve for the electromagnetic modes of a system defined by in-plane permittivity and
        permeability and assuming translational invariance in the normal direction.

        Parameters
        ----------
        Nx : int
            Number of grids along x-direction.
        Ny : int
            Number of grids along y-direction.
        eps_tensor : np.ndarray
            Shape (3, 3, N), the permittivity tensor at every point in the plane.
        mu_tensor : np.ndarray
            Shape (3, 3, N), the permittivity tensor at every point in the plane.
        der_mats : List[scipy.sparse.csr_matrix]
            The sparse derivative matrices dxf, dxb, dyf, dyb, including the PML.
        num_modes : int
            Number of modes to solve for.
        neff_guess : float
            Initial guess for the effective index.
        mat_precision : Union['single', 'double']
            Single or double-point precision in eigensolver.
        direction : Union["+", "-"]
            Direction of mode propagation.
        basis_E: np.ndarray
            Basis for mode solving.
        dls: Tuple[List[np.ndarray], List[np.ndarray]]
            Primal and dual grid steps along each of the two tangential dimensions.
        dmin_pmc: List[bool]
            List of booleans indicating whether to apply PMC at the near end of the grid.

        Returns
        -------
        E : np.ndarray
            Electric field of the eigenmodes, shape (3, N, num_modes).
        H : np.ndarray
            Magnetic field of the eigenmodes, shape (3, N, num_modes).
        neff : np.ndarray
            Real part of the effective index, shape (num_modes, ).
        keff : np.ndarray
            Imaginary part of the effective index, shape (num_modes, ).
        eps_spec : Union["diagonal", "tensorial_real", "tensorial_complex"]
            Permittivity characterization on the mode solver's plane.
        """

        # In the matrices P and Q below, they contain terms ``epsilon_parallel`` or
        # ``mu_parallel``, and also a term proportional to 1/(k0 * dl)**2. To make sure
        # that permittivity of good conductor is visible in low-frequency/high resolution, pec_val should be
        # scaled by a factor max(1, max(1/k0 dl) **2).
        pec_scaling = max(1, max([np.max(abs(f)) for f in der_mats]) ** 2)
        pec_scaled_val = min(GOOD_CONDUCTOR_CUT_OFF, pec_scaling * abs(pec_val))

        # use a high-conductivity model for locations associated with a good conductor
        def conductivity_model_for_good_conductor(
            eps, threshold=GOOD_CONDUCTOR_THRESHOLD * pec_val
        ):
            """Entries associated with 'eps' are converted to a high-conductivity model."""
            eps = eps.astype(complex)
            eps[np.abs(eps) >= abs(threshold)] = 1 + 1j * pec_scaled_val
            return eps

        eps_tensor = conductivity_model_for_good_conductor(eps_tensor)
        mu_tensor = conductivity_model_for_good_conductor(mu_tensor)

        # Determine if ``eps`` and ``mu`` are diagonal or tensorial
        off_diagonals = (np.ones((3, 3)) - np.eye(3)).astype(bool)
        eps_offd = np.abs(eps_tensor[off_diagonals])
        mu_offd = np.abs(mu_tensor[off_diagonals])
        is_tensorial = np.any(eps_offd > TOL_TENSORIAL) or np.any(mu_offd > TOL_TENSORIAL)

        # Determine if ``eps`` and ``mu`` represent reciprocal media
        is_reciprocal = True
        if is_tensorial:
            is_reciprocal = cls._check_reciprocity(
                eps_tensor, TOL_TENSORIAL
            ) and cls._check_reciprocity(mu_tensor, TOL_TENSORIAL)

        # initial vector for eigensolver in correct data type
        vec_init = cls.set_initial_vec(Nx, Ny, is_tensorial=is_tensorial)

        # call solver: base kwargs shared by diagonal and tensorial solvers
        base_kwargs = {
            "eps": eps_tensor,
            "mu": mu_tensor,
            "der_mats": der_mats,
            "num_modes": num_modes,
            "neff_guess": neff_guess,
            "vec_init": vec_init,
            "mat_precision": mat_precision,
        }

        if basis_E is not None and is_tensorial:
            raise RuntimeError(
                "Tensorial eps not yet supported in relative mode solver "
                "(with basis fields provided)."
            )

        # Determine if epsilon has complex values (used to select real vs complex tensorial solver)
        is_eps_complex = cls.isinstance_complex(eps_tensor)

        if not is_tensorial:
            eps_spec = "diagonal"
            E, H, neff, keff = cls.solver_diagonal(
                **base_kwargs,
                enable_incidence_matrices=enable_incidence_matrices,
                basis_E=basis_E,
            )
            if direction == "-":
                H[0] *= -1
                H[1] *= -1
                E[2] *= -1

        elif not is_eps_complex:
            eps_spec = "tensorial_real"
            E, H, neff, keff = cls.solver_tensorial(
                **base_kwargs,
                direction="+",
                dls=dls,
                Nxy=(Nx, Ny),
                dmin_pmc=dmin_pmc,
            )
            if direction == "-":
                E = np.conj(E)
                H = -np.conj(H)

        else:
            eps_spec = "tensorial_complex"
            E, H, neff, keff = cls.solver_tensorial(
                **base_kwargs,
                direction=direction,
                dls=dls,
                Nxy=(Nx, Ny),
                dmin_pmc=dmin_pmc,
            )

        # We only can orthogonalize reciprocal modes in this manner.
        if is_reciprocal:
            dl_f, dl_b = dls
            # Normalize all modes so self-dot equals 1.0 and find any self-orthogonal modes
            all_modes = list(range(num_modes))
            original_shape = E.shape
            E = E.reshape((3, Nx, Ny, num_modes))
            H = H.reshape((3, Nx, Ny, num_modes))
            E, H, self_orthogonal_modes = cls._normalize_modes(
                E, H, dl_f, dl_b, all_modes, TOL_NEEDS_ORTHOGONALIZATION
            )
            # Identify and orthogonalize degenerate modes, excluding self-orthogonal modes
            E, H = cls._orthogonalize_degenerate_modes(
                E, H, neff, keff, dl_f, dl_b, self_orthogonal_modes
            )
            E = E.reshape(original_shape)
            H = H.reshape(original_shape)

        return E, H, neff, keff, eps_spec

    @classmethod
    def solver_diagonal(
        cls,
        eps,
        mu,
        der_mats,
        num_modes,
        neff_guess,
        vec_init,
        mat_precision,
        enable_incidence_matrices,
        basis_E,
    ):
        """EM eigenmode solver assuming ``eps`` and ``mu`` are diagonal everywhere."""
        import scipy.sparse as sp
        import scipy.sparse.linalg as spl

        # code associated with these options is included below in case it's useful in the future
        enable_preconditioner = False
        analyze_conditioning = False
        _threshold = 0.9 * np.abs(pec_val)

        def incidence_matrix_for_pec(eps_vec, threshold=_threshold):
            """Incidence matrix indicating non-PEC entries associated with 'eps_vec'."""
            nnz = eps_vec[np.abs(eps_vec) < threshold]
            eps_nz = eps_vec.copy()
            eps_nz[np.abs(eps_vec) >= threshold] = 0
            rows = np.arange(0, len(nnz))
            cols = np.argwhere(eps_nz).flatten()
            dnz = sp.csr_matrix(([1] * len(nnz), (rows, cols)), shape=(len(rows), len(eps_vec)))
            return dnz

        mode_solver_type = "diagonal"
        N = eps.shape[-1]

        # Unpack eps, mu and derivatives
        eps_xx = eps[0, 0, :]
        eps_yy = eps[1, 1, :]
        eps_zz = eps[2, 2, :]
        mu_xx = mu[0, 0, :]
        mu_yy = mu[1, 1, :]
        mu_zz = mu[2, 2, :]
        dxf, dxb, dyf, dyb = der_mats

        if any(
            cls.mode_plane_contain_good_conductor(i)
            for i in [eps_xx, eps_yy, eps_zz, mu_xx, mu_yy, mu_zz]
        ):
            enable_preconditioner = True

        # Compute the matrix for diagonalization
        inv_eps_zz = sp.spdiags(1 / eps_zz, [0], N, N)
        inv_mu_zz = sp.spdiags(1 / mu_zz, [0], N, N)

        if enable_incidence_matrices:
            dnz_xx, dnz_yy, dnz_zz = (incidence_matrix_for_pec(i) for i in [eps_xx, eps_yy, eps_zz])
            dnz = sp.block_diag((dnz_xx, dnz_yy), format="csr")
            inv_eps_zz = (dnz_zz.T) * dnz_zz * inv_eps_zz * (dnz_zz.T) * dnz_zz

        # P = p_mu + p_partial
        # Q = q_ep + q_partial
        # Note that p_partial @ q_partial = 0, so that PQ = p_mu @ Q + p_partial @ q_ep
        p_mu = sp.bmat(
            [[None, sp.spdiags(mu_yy, [0], N, N)], [sp.spdiags(-mu_xx, [0], N, N), None]]
        )
        p_partial = sp.bmat(
            [
                [-dxf.dot(inv_eps_zz).dot(dyb), dxf.dot(inv_eps_zz).dot(dxb)],
                [-dyf.dot(inv_eps_zz).dot(dyb), dyf.dot(inv_eps_zz).dot(dxb)],
            ]
        )
        q_ep = sp.bmat(
            [[None, sp.spdiags(eps_yy, [0], N, N)], [sp.spdiags(-eps_xx, [0], N, N), None]]
        )
        q_partial = sp.bmat(
            [
                [-dxb.dot(inv_mu_zz).dot(dyf), dxb.dot(inv_mu_zz).dot(dxf)],
                [-dyb.dot(inv_mu_zz).dot(dyf), dyb.dot(inv_mu_zz).dot(dxf)],
            ]
        )

        # pmat = p_mu + p_partial  # no need to assemble pmat, as it is not used anywhere
        qmat = q_ep + q_partial
        mat = p_mu @ qmat + p_partial @ q_ep

        # Cast matrix to target data type
        mat_dtype = cls.matrix_data_type(eps, mu, der_mats, mat_precision, is_tensorial=False)
        mat = cls.type_conversion(mat, mat_dtype)

        # Casting starting vector to target data type
        vec_init = cls.type_conversion(vec_init, mat_dtype)

        # Starting eigenvalue guess in target data type
        eig_guess = cls.type_conversion(np.array([-(neff_guess**2)]), mat_dtype)[0]

        if enable_incidence_matrices:
            mat = dnz * mat * dnz.T
            vec_init = dnz * vec_init

        # Denote the original eigenvalue problem as Ax = lambda x,
        # with left and right preconditioners, we solve for the following eigenvalue problem,
        # L A R y = lambda LR y, where x = R y
        precon_left = None
        precon_right = None
        generalized_M = None  # matrix in the generalized eigenvalue problem
        if enable_preconditioner:
            if PRECONDITIONER == "Jacobi":
                precon_right = sp.diags(1 / mat.diagonal())

            elif PRECONDITIONER == "Material":

                def conditional_inverted_vec(eps_vec, threshold=1):
                    """Returns a diagonal sparse matrix whose i-th element in the diagonal
                    is |eps_i|^-1 if |eps_i|>threshold, and |eps_i| otherwise.
                    """
                    abs_vec = np.abs(eps_vec)
                    return sp.spdiags(
                        np.where(abs_vec > threshold, 1.0 / abs_vec, abs_vec), [0], N, N
                    )

                precon_left = sp.bmat(
                    [
                        [conditional_inverted_vec(mu_yy), None],
                        [None, conditional_inverted_vec(mu_xx)],
                    ]
                )
                precon_right = sp.bmat(
                    [
                        [conditional_inverted_vec(eps_xx), None],
                        [None, conditional_inverted_vec(eps_yy)],
                    ]
                )
            generalized_M = precon_right
            mat = mat @ precon_right
            if precon_left is not None:
                generalized_M = precon_left @ generalized_M
                mat = precon_left @ mat

        # Trim small values in single precision case
        if mat_precision == "single":
            cls.trim_small_values(mat, tol=fp_eps)

        if analyze_conditioning:
            aca = mat.conjugate().T * mat
            aac = mat * mat.conjugate().T
            diff = aca - aac
            print(
                f"inf-norm: A*A: {spl.norm(aca, ord=np.inf)}, AA*: {spl.norm(aac, ord=np.inf)}, nonnormality: {spl.norm(diff, ord=np.inf)}, relative nonnormality: {spl.norm(diff, ord=np.inf) / spl.norm(aca, ord=np.inf)}"
            )
            print(
                f"fro-norm: A*A: {spl.norm(aca, ord='fro')}, AA*: {spl.norm(aac, ord='fro')}, nonnormality: {spl.norm(diff, ord='fro')}, relative nonnormality: {spl.norm(diff, ord='fro') / spl.norm(aca, ord='fro')}"
            )

        # preprocess basis modes
        basis_vecs = None
        if basis_E is not None:
            basis_Ex = basis_E[0, ...]
            basis_Ey = basis_E[1, ...]
            basis_vecs = np.concatenate((basis_Ex, basis_Ey), axis=0)

            # if enable_preconditioner:
            #    basis_vecs = (1 / precon) * basis_vecs

            # if enable_incidence_matrices:
            #    basis_vecs = dnz * basis_vecs

        # Call the eigensolver. The eigenvalues are -(neff + 1j * keff)**2
        if basis_E is None:
            vals, vecs = cls.solver_eigs(
                mat,
                num_modes,
                vec_init,
                guess_value=eig_guess,
                mode_solver_type=mode_solver_type,
                M=generalized_M,
            )
        else:
            vals, vecs = cls.solver_eigs_relative(
                mat,
                num_modes,
                vec_init,
                guess_value=eig_guess,
                mode_solver_type=mode_solver_type,
                M=generalized_M,
                basis_vecs=basis_vecs,
            )
        neff, keff = cls.eigs_to_effective_index(vals, mode_solver_type)

        # Sort by descending neff
        sort_inds = np.argsort(neff)[::-1]
        neff = neff[sort_inds]
        keff = keff[sort_inds]

        E, H = None, None
        if basis_E is None:
            if precon_right is not None:
                vecs = precon_right * vecs

            if enable_incidence_matrices:
                vecs = dnz.T * vecs

        vecs = vecs[:, sort_inds]

        # Field components from eigenvectors
        Ex = vecs[:N, :]
        Ey = vecs[N:, :]

        # Get the other field components
        h_field = qmat.dot(vecs)
        Hx = h_field[:N, :] / (1j * neff - keff)
        Hy = h_field[N:, :] / (1j * neff - keff)
        Hz = inv_mu_zz.dot(dxf.dot(Ey) - dyf.dot(Ex))
        # Ez = inv_eps_zz.dot(dxb.dot(Hy) - dyb.dot(Hx))

        # Ez = -inv_eps_zz * div^H J H_xy, while Hxy = vals^-1 * qmat * Exy;
        # Note that div^H J q_partial = 0, so Ez = -vals^-1 inv_eps_zz * div^H J q_ep Exy
        h_partial_field = q_ep.dot(vecs) / (1j * neff - keff)
        Ez = inv_eps_zz.dot(dxb.dot(h_partial_field[N:, :]) - dyb.dot(h_partial_field[:N, :]))

        # Bundle up
        E = np.stack((Ex, Ey, Ez), axis=0)
        H = np.stack((Hx, Hy, Hz), axis=0)

        # Return to standard H field units (see CEM notes for H normalization used in solver)
        H *= -1j / ETA_0

        return E, H, neff, keff

    @classmethod
    def matrix_data_type(cls, eps, mu, der_mats, mat_precision, is_tensorial):
        """Determine data type that should be used for the matrix for diagonalization."""
        mat_dtype = np.float32
        # In tensorial case, even though the matrix can be real, the
        # expected eigenvalue is purely imaginary. So for now we enforce
        # the matrix to be complex type so that it will look for the right eigenvalues.
        if is_tensorial:
            mat_dtype = np.complex128 if mat_precision == "double" else np.complex64
        else:
            # 1) check if complex or not
            complex_solver = (
                cls.isinstance_complex(eps)
                or cls.isinstance_complex(mu)
                or np.any([cls.isinstance_complex(f) for f in der_mats])
            )
            # 2) determine precision
            if complex_solver:
                mat_dtype = np.complex128 if mat_precision == "double" else np.complex64
            else:
                if mat_precision == "double":
                    mat_dtype = np.float64

        return mat_dtype

    @classmethod
    def trim_small_values(cls, mat: sp.csr_matrix, tol: float) -> sp.csr_matrix:
        """Eliminate elements of matrix ``mat`` for which ``abs(element) / abs(max_element) < tol``,
        or ``np.abs(mat_data) < tol``. This operates in-place on mat so there is no return.
        """
        max_element = np.amax(np.abs(mat))
        mat.data *= np.logical_or(np.abs(mat.data) / max_element > tol, np.abs(mat.data) > tol)
        mat.eliminate_zeros()
        return mat

    @staticmethod
    def _check_reciprocity(material_tensor: np.ndarray, tol: float) -> bool:
        """Check if material tensor is symmetric (reciprocal): tensor[i,j] ≈ tensor[j,i]."""
        diff = np.abs(material_tensor - material_tensor.transpose(1, 0, 2))
        max_error = np.max(diff)
        return max_error <= tol

    @classmethod
    def solver_tensorial(
        cls,
        eps,
        mu,
        der_mats,
        num_modes,
        neff_guess,
        vec_init,
        mat_precision,
        direction,
        dls,
        Nxy=None,
        dmin_pmc=None,
    ):
        """EM eigenmode solver assuming ``eps`` or ``mu`` have off-diagonal elements."""
        import scipy.sparse as sp

        mode_solver_type = "tensorial"
        N = eps.shape[-1]
        dxf, dxb, dyf, dyb = der_mats

        # Compute all blocks of the matrix for diagonalization
        inv_eps_zz = sp.spdiags(1 / eps[2, 2, :], [0], N, N)
        inv_mu_zz = sp.spdiags(1 / mu[2, 2, :], [0], N, N)
        axax = -dxf.dot(sp.spdiags(eps[2, 0, :] / eps[2, 2, :], [0], N, N)) - sp.spdiags(
            mu[1, 2, :] / mu[2, 2, :], [0], N, N
        ).dot(dyf)
        axay = -dxf.dot(sp.spdiags(eps[2, 1, :] / eps[2, 2, :], [0], N, N)) + sp.spdiags(
            mu[1, 2, :] / mu[2, 2, :], [0], N, N
        ).dot(dxf)
        axbx = -dxf.dot(inv_eps_zz).dot(dyb) + sp.spdiags(
            mu[1, 0, :] - mu[1, 2, :] * mu[2, 0, :] / mu[2, 2, :], [0], N, N
        )
        axby = dxf.dot(inv_eps_zz).dot(dxb) + sp.spdiags(
            mu[1, 1, :] - mu[1, 2, :] * mu[2, 1, :] / mu[2, 2, :], [0], N, N
        )
        ayax = -dyf.dot(sp.spdiags(eps[2, 0, :] / eps[2, 2, :], [0], N, N)) + sp.spdiags(
            mu[0, 2, :] / mu[2, 2, :], [0], N, N
        ).dot(dyf)
        ayay = -dyf.dot(sp.spdiags(eps[2, 1, :] / eps[2, 2, :], [0], N, N)) - sp.spdiags(
            mu[0, 2, :] / mu[2, 2, :], [0], N, N
        ).dot(dxf)
        aybx = -dyf.dot(inv_eps_zz).dot(dyb) + sp.spdiags(
            -mu[0, 0, :] + mu[0, 2, :] * mu[2, 0, :] / mu[2, 2, :], [0], N, N
        )
        ayby = dyf.dot(inv_eps_zz).dot(dxb) + sp.spdiags(
            -mu[0, 1, :] + mu[0, 2, :] * mu[2, 1, :] / mu[2, 2, :], [0], N, N
        )
        bxbx = -dxb.dot(sp.spdiags(mu[2, 0, :] / mu[2, 2, :], [0], N, N)) - sp.spdiags(
            eps[1, 2, :] / eps[2, 2, :], [0], N, N
        ).dot(dyb)
        bxby = -dxb.dot(sp.spdiags(mu[2, 1, :] / mu[2, 2, :], [0], N, N)) + sp.spdiags(
            eps[1, 2, :] / eps[2, 2, :], [0], N, N
        ).dot(dxb)
        bxax = -dxb.dot(inv_mu_zz).dot(dyf) + sp.spdiags(
            eps[1, 0, :] - eps[1, 2, :] * eps[2, 0, :] / eps[2, 2, :], [0], N, N
        )
        bxay = dxb.dot(inv_mu_zz).dot(dxf) + sp.spdiags(
            eps[1, 1, :] - eps[1, 2, :] * eps[2, 1, :] / eps[2, 2, :], [0], N, N
        )
        bybx = -dyb.dot(sp.spdiags(mu[2, 0, :] / mu[2, 2, :], [0], N, N)) + sp.spdiags(
            eps[0, 2, :] / eps[2, 2, :], [0], N, N
        ).dot(dyb)
        byby = -dyb.dot(sp.spdiags(mu[2, 1, :] / mu[2, 2, :], [0], N, N)) - sp.spdiags(
            eps[0, 2, :] / eps[2, 2, :], [0], N, N
        ).dot(dxb)
        byax = -dyb.dot(inv_mu_zz).dot(dyf) + sp.spdiags(
            -eps[0, 0, :] + eps[0, 2, :] * eps[2, 0, :] / eps[2, 2, :], [0], N, N
        )
        byay = dyb.dot(inv_mu_zz).dot(dxf) + sp.spdiags(
            -eps[0, 1, :] + eps[0, 2, :] * eps[2, 1, :] / eps[2, 2, :], [0], N, N
        )

        mat = sp.bmat(
            [
                [axax, axay, axbx, axby],
                [ayax, ayay, aybx, ayby],
                [bxax, bxay, bxbx, bxby],
                [byax, byay, bybx, byby],
            ]
        )

        # The eigenvalues for the matrix above are 1j * (neff + 1j * keff)
        # Multiply the matrix by -1j, so that eigenvalues are (neff + 1j * keff)
        mat *= -1j

        # change matrix sign for backward direction
        if direction == "-":
            mat *= -1

        # Cast matrix to target data type
        mat_dtype = cls.matrix_data_type(eps, mu, der_mats, mat_precision, is_tensorial=True)
        mat = cls.type_conversion(mat, mat_dtype)

        # Trim small values in single precision case
        if mat_precision == "single":
            cls.trim_small_values(mat, tol=fp_eps)

        # Casting starting vector to target data type
        vec_init = cls.type_conversion(vec_init, mat_dtype)

        # Starting eigenvalue guess in target data type
        eig_guess = cls.type_conversion(np.array([neff_guess]), mat_dtype)[0]

        # Call the eigensolver.
        vals, vecs = cls.solver_eigs(
            mat,
            num_modes,
            vec_init,
            guess_value=eig_guess,
            mode_solver_type=mode_solver_type,
        )
        neff, keff = cls.eigs_to_effective_index(vals, mode_solver_type)
        # Sort by descending real part
        sort_inds = np.argsort(neff)[::-1]
        neff = neff[sort_inds]
        keff = keff[sort_inds]
        vecs = vecs[:, sort_inds]

        # Field components from eigenvectors
        Ex = vecs[:N, :]
        Ey = vecs[N : 2 * N, :]
        Hx = vecs[2 * N : 3 * N, :]
        Hy = vecs[3 * N :, :]

        # Get the other field components
        hxy_term = (-mu[2, 0, :] * Hx.T - mu[2, 1, :] * Hy.T).T
        Hz = inv_mu_zz.dot(dxf.dot(Ey) - dyf.dot(Ex) + hxy_term)
        exy_term = (-eps[2, 0, :] * Ex.T - eps[2, 1, :] * Ey.T).T
        Ez = inv_eps_zz.dot(dxb.dot(Hy) - dyb.dot(Hx) + exy_term)

        # Bundle up
        E = np.stack((Ex, Ey, Ez), axis=0)
        H = np.stack((Hx, Hy, Hz), axis=0)

        # Return to standard H field units (see CEM notes for H normalization used in solver)
        # The minus sign here is suspicious, need to check how modes are used in Mode objects
        H *= -1j / ETA_0

        return E, H, neff, keff

    @classmethod
    def solver_eigs(
        cls,
        mat,
        num_modes,
        vec_init,
        guess_value=1.0,
        M=None,
        **kwargs: Any,
    ):
        """Find ``num_modes`` eigenmodes of ``mat`` cloest to ``guess_value``.

        Parameters
        ----------
        mat : scipy.sparse matrix
            Square matrix for diagonalization.
        num_modes : int
            Number of eigenmodes to compute.
        guess_value : float, optional
        """
        import scipy.sparse.linalg as spl

        values, vectors = spl.eigs(
            mat, k=num_modes, sigma=guess_value, tol=TOL_EIGS, v0=vec_init, M=M
        )

        # for i, eig_i in enumerate(values):
        #     vec = vectors[:, i]
        #     rhs = vec
        #     if M is not None:
        #         rhs = M @ rhs
        #     eig_from_vec = (vec.T @ (mat @ vec)) / (vec.T @ rhs)
        #     residue = np.linalg.norm(mat @ vec - eig_i * rhs) / np.linalg.norm(vec)
        #     print(
        #         f"{i}-th eigenvalue: {eig_i}, referred from eigenvectors: {eig_from_vec}, relative residual: {residue}."
        #     )
        return values, vectors

    @classmethod
    def solver_eigs_relative(
        cls,
        mat,
        num_modes,
        vec_init,
        guess_value=1.0,
        M=None,
        basis_vecs=None,
        **kwargs: Any,
    ):
        """Find ``num_modes`` eigenmodes of ``mat`` cloest to ``guess_value``.

        Parameters
        ----------
        mat : scipy.sparse matrix
            Square matrix for diagonalization.
        num_modes : int
            Number of eigenmodes to compute.
        guess_value : float, optional
        """
        import scipy.linalg as linalg

        basis, _ = np.linalg.qr(basis_vecs)
        mat_basis = np.conj(basis.T) @ mat @ basis
        values, coeffs = linalg.eig(mat_basis)
        vectors = None
        vectors = basis @ coeffs
        return values, vectors

    @classmethod
    def isinstance_complex(cls, vec_or_mat, tol=TOL_COMPLEX):
        """Check if a numpy array or scipy.sparse.csr_matrix has complex component by looking at
        norm(x.imag)/norm(x)>TOL_COMPLEX

        Parameters
        ----------
        vec_or_mat : Union[np.ndarray, sp.csr_matrix]
        """
        import scipy.sparse.linalg as spl
        from scipy.sparse import csr_matrix

        if isinstance(vec_or_mat, np.ndarray):
            return np.linalg.norm(vec_or_mat.imag) / (np.linalg.norm(vec_or_mat) + fp_eps) > tol
        if isinstance(vec_or_mat, csr_matrix):
            mat_norm = spl.norm(vec_or_mat)
            mat_imag_norm = spl.norm(vec_or_mat.imag)
            return mat_imag_norm / (mat_norm + fp_eps) > tol
        raise RuntimeError(
            f"Variable type should be either numpy array or scipy.sparse.csr_matrix, got {type(vec_or_mat)}."
        )

    @classmethod
    def type_conversion(cls, vec_or_mat, new_dtype):
        """Convert vec_or_mat to new_type.

        Parameters
        ----------
        vec_or_mat : Union[np.ndarray, sp.csr_matrix]
            vec or mat to be converted.
        new_dtype : Union[np.complex128, np.complex64, np.float64, np.float32]
            Final type of vec or mat

        Returns
        -------
        converted_vec_or_mat : Union[np.ndarray, sp.csr_matrix]
        """

        if new_dtype in {np.complex128, np.complex64}:
            return vec_or_mat.astype(new_dtype)
        if new_dtype in {np.float64, np.float32}:
            converted_vec_or_mat = vec_or_mat.real
            return converted_vec_or_mat.astype(new_dtype)

        raise RuntimeError("Unsupported new_type.")

    @classmethod
    def set_initial_vec(cls, Nx, Ny, is_tensorial=False):
        """Set initial vector for eigs:
        1) The field at x=0 and y=0 boundaries are set to 0. This should be
        the case for PEC boundaries, but wouldn't hurt for non-PEC boundary;
        2) The vector is np.complex128 by default, and will be converted to
        appropriate type afterwards.

        Parameters
        ----------
        Nx : int
            Number of grids along x-direction.
        Ny : int
            Number of grids along y-direction.
        is_tensorial : bool
            diagonal or tensorial eigenvalue problem.
        """

        # The size of the vector is len_multiplier * Nx * Ny
        len_multiplier = 2
        if is_tensorial:
            len_multiplier *= 2

        # Initialize the vector
        size = (Nx, Ny, len_multiplier)
        rng = np.random.default_rng(0)
        vec_init = rng.random(size) + 1j * rng.random(size)

        # Set values at the boundary to be 0
        if Nx > 1:
            vec_init[0, :, :] = 0
        if Ny > 1:
            vec_init[:, 0, :] = 0

        # Concatenate the vector appropriately
        vec_init = np.vstack(vec_init)
        return vec_init.flatten("F")

    @classmethod
    def eigs_to_effective_index(cls, eig_list: Numpy, mode_solver_type: ModeSolverType):
        """Convert obtained eigenvalues to n_eff and k_eff.

        Parameters
        ----------
        eig_list : Numpy
            Array of eigenvalues
        mode_solver_type : ModeSolverType
            The type of mode solver problems

        Returns
        -------
        Tuple[Numpy, Numpy]
            n_eff and k_eff
        """
        if eig_list.size == 0:
            raise RuntimeError("Could not find any eigenmodes for this waveguide.")

        # for tensorial type, it's simply (neff + 1j * keff)
        if mode_solver_type == "tensorial":
            return np.real(eig_list), np.imag(eig_list)

        # for diagonal type, eigenvalues are -(neff + 1j * keff)**2
        if mode_solver_type == "diagonal":
            sqrt_eig_list = np.emath.sqrt(-eig_list + 0j)
            return np.real(sqrt_eig_list), np.imag(sqrt_eig_list)

        raise RuntimeError(f"Unidentified 'mode_solver_type={mode_solver_type}'.")

    @staticmethod
    def format_medium_data(mat_data):
        """
        mat_data can be either permittivity or permeability. It's either a single 2D array
        defining the relative property in the cross-section, or nine 2D arrays defining
        the property at the E(H)x, E(H)y, and E(H)z locations of the Yee grid in the order
        xx, xy, xz, yx, yy, yz, zx, zy, zz.
        """
        if isinstance(mat_data, Numpy):
            return (mat_data[i, :, :] for i in range(9))
        if len(mat_data) == 9:
            return (np.copy(e) for e in mat_data)
        raise ValueError("Wrong input to mode solver pemittivity/permeability!")

    @staticmethod
    def split_curl_field_postprocess(split_curl, E):
        """E has the shape (3, N, num_modes)"""
        _, Nx, Ny = split_curl.shape
        field_shape = E.shape

        # set a dummy value of split curl inside PEC to avoid division by 0 warning (it's 0/0, since
        # E field inside PEC is also 0); then by the end, zero out E inside PEC again just to be safe.
        outside_pec = ~np.isclose(split_curl, 0)
        split_curl_scaling = np.copy(split_curl)
        split_curl_scaling[~outside_pec] = 1.0

        E = E.reshape(3, Nx, Ny, field_shape[-1])
        E /= split_curl_scaling[:, :, :, np.newaxis]
        E *= outside_pec[:, :, :, np.newaxis]
        E = E.reshape(field_shape)
        return E

    @staticmethod
    def make_pml_invariant(Nxy, tensor, num_pml):
        """For a given epsilon or mu tensor of shape ``(3, 3, Nx, Ny)``, and ``num_pml`` pml layers
        along ``x`` and ``y``, make all the tensor values in the PML equal by replicating the first
        pixel into the PML."""

        Nx, Ny = Nxy
        new_ten = tensor.reshape((3, 3, Nx, Ny))
        new_ten[:, :, : num_pml[0], :] = new_ten[:, :, num_pml[0], :][:, :, None, :]
        new_ten[:, :, Nx - num_pml[0] + 1 :, :] = new_ten[:, :, Nx - num_pml[0], :][:, :, None, :]
        new_ten[:, :, :, : num_pml[1]] = new_ten[:, :, :, num_pml[1]][:, :, :, None]
        new_ten[:, :, :, Ny - num_pml[1] + 1 :] = new_ten[:, :, :, Ny - num_pml[1]][:, :, :, None]
        return new_ten.reshape((3, 3, -1))

    @staticmethod
    def split_curl_field_postprocess_inverse(split_curl, E) -> None:
        """E has the shape (3, N, num_modes)"""
        raise RuntimeError("Split curl not yet implemented for relative mode solver.")

    @staticmethod
    def mode_plane_contain_good_conductor(material_response) -> bool:
        """Find out if epsilon on the modal plane contain good conductors whose permittivity
        or permeability value is very large.
        """
        if material_response is None:
            return False
        return np.any(np.abs(material_response) > GOOD_CONDUCTOR_THRESHOLD * np.abs(pec_val))

    @staticmethod
    def _identify_degenerate_modes(
        n_complex: np.ndarray,
        tol: float,
    ) -> list[set[int]]:
        """Inspects the n_complex of modes to find sets of degenerate modes."""
        num_modes = len(n_complex)
        ungrouped = set(range(num_modes))
        degenerate_sets = []

        while ungrouped:
            # Start a new group with an ungrouped mode
            seed = ungrouped.pop()
            current_set = {seed}

            # Find all ungrouped modes similar to the seed
            to_remove = set()
            for mode_idx in ungrouped:
                if np.isclose(n_complex[mode_idx], n_complex[seed], rtol=tol, atol=tol):
                    current_set.add(mode_idx)
                    to_remove.add(mode_idx)

            # Remove grouped modes from ungrouped set
            ungrouped -= to_remove

            # Only keep groups with more than one mode
            if len(current_set) >= 2:
                degenerate_sets.append(current_set)

        return degenerate_sets

    @staticmethod
    def _identify_degenerate_modes_with_dot(
        degenerate_sets: list[set[int]],
        E: np.ndarray,
        H: np.ndarray,
        dl_primal: np.ndarray,
        dl_dual: np.ndarray,
        tol: float,
    ) -> list[set[int]]:
        """Inspects the biorthogonality condition between modes to filter sets of degenerate modes.

        For each set of modes identified as degenerate by neff similarity, this function computes
        the overlap matrix and uses connected components to find which modes are truly degenerate
        (have significant mutual overlap).

        Returns only groups with >= 2 modes that have significant overlap.
        """
        import scipy.sparse as sp

        filtered_degenerate_sets = []

        for degenerate_group in degenerate_sets:
            # Convert set to list for indexing
            mode_indices = sorted(degenerate_group)

            # S is a symmetric overlap matrix of the degenerate modes
            S = EigSolver._outer_dot(E, H, dl_primal, dl_dual, mode_indices)

            # Create adjacency matrix
            # True if overlap is significant, False if orthogonal
            # Zero out diagonal so a mode doesn't connect to itself
            adjacency = (np.abs(S) > tol).astype(int)
            np.fill_diagonal(adjacency, 0)

            # Find connected components
            # labels: array indicating which group each mode belongs to
            _, labels = sp.csgraph.connected_components(sp.csr_matrix(adjacency), directed=False)

            # Group modes by their connected component label
            groups = {}
            for label_idx, group_id in enumerate(labels):
                mode_index = mode_indices[label_idx]
                if group_id not in groups:
                    groups[group_id] = []
                groups[group_id].append(mode_index)

            # Convert groups dict to list of sets, keeping only groups with >= 2 modes
            for group_modes in groups.values():
                if len(group_modes) >= 2:
                    filtered_degenerate_sets.append(set(group_modes))

        return filtered_degenerate_sets

    @staticmethod
    def _dot(
        E: np.ndarray,
        H: np.ndarray,
        dl_primal: np.ndarray,
        dl_dual: np.ndarray,
        mode_1: int,
        mode_2: int,
    ) -> complex:
        """Dot product based on the bi-orthogonality relationship between E and H."""
        # Extract field components
        Ex = E[0, ...]
        Ey = E[1, ...]
        Hx = H[0, ...]
        Hy = H[1, ...]

        # Make the differential area elements
        Ex_Hy_dS = np.outer(dl_primal[0], dl_dual[1])
        Ey_Hx_dS = np.outer(dl_dual[0], dl_primal[1])

        term1 = Ex[..., mode_1] * Hy[..., mode_2] + Ex[..., mode_2] * Hy[..., mode_1]
        term1 *= Ex_Hy_dS
        term2 = Ey[..., mode_1] * Hx[..., mode_2] + Ey[..., mode_2] * Hx[..., mode_1]
        term2 *= Ey_Hx_dS
        return (1 / 4) * np.sum(term1 - term2)

    @staticmethod
    def _cauchy_schwarz_dot_bound(
        E: np.ndarray,
        H: np.ndarray,
        dl_primal: np.ndarray,
        dl_dual: np.ndarray,
        mode_idx: int,
    ) -> complex:
        """Calculates the upper bound for the self-dot (overlap) using the Cauchy-Schwarz inequality.

        Hard to predict how the normalization of the eigenvectors from the linear algebra solver
        translates to normalization in terms of the EM overlap integral,
        so we use this upper bound to determine whether a mode should be considered as self-orthogonal.

        Sketch of derivation for the upper bound:
        | S (E x H) . z dA |               (Absolute value of the bi-orthogonality overlap integral)
        <=  S |(E x H) . z| dA             (Triangle Inequality)
        <=  S |E| * |H|     dA             (Cross Product Magnitude)
        <= sqrt(S |E|^2) * sqrt(S |H|^2)   (Cauchy-Schwarz)
        """
        # Extract all field components
        Ex = E[0, ...]
        Ey = E[1, ...]
        Ez = E[2, ...]
        Hx = H[0, ...]
        Hy = H[1, ...]
        Hz = H[2, ...]

        # Make the differential area elements
        Ex_Hy_dS = np.outer(dl_primal[0], dl_dual[1])
        Ey_Hx_dS = np.outer(dl_dual[0], dl_primal[1])
        Ez_dS = np.outer(dl_dual[0], dl_dual[1])
        Hz_dS = np.outer(dl_primal[0], dl_primal[1])

        E_int = np.sum(Ex[..., mode_idx] * Ex[..., mode_idx].conj() * Ex_Hy_dS)
        E_int += np.sum(Ey[..., mode_idx] * Ey[..., mode_idx].conj() * Ey_Hx_dS)
        E_int += np.sum(Ez[..., mode_idx] * Ez[..., mode_idx].conj() * Ez_dS)
        E_int = E_int / 2

        H_int = np.sum(Hx[..., mode_idx] * Hx[..., mode_idx].conj() * Ey_Hx_dS)
        H_int += np.sum(Hy[..., mode_idx] * Hy[..., mode_idx].conj() * Ex_Hy_dS)
        H_int += np.sum(Hz[..., mode_idx] * Hz[..., mode_idx].conj() * Hz_dS)
        H_int = H_int / 2

        return np.sqrt(E_int * H_int)

    @staticmethod
    def _outer_dot(
        E: np.ndarray,
        H: np.ndarray,
        dl_primal: np.ndarray,
        dl_dual: np.ndarray,
        mode_indices: set[int],
    ) -> np.ndarray:
        """Vectorized modal overlap matrix calculation for a set of modes.

        This overlap is based on the bi-orthogonality relationship between E and H.
        Returns a matrix S where S[i, j] is the overlap between mode i and mode j.

        Note: The overlap matrix is symmetric, so we only compute the upper triangle
        (including diagonal) and fill the lower triangle by symmetry.
        """
        # Convert set to sorted list for consistent indexing
        mode_list = sorted(mode_indices)
        n_modes = len(mode_list)

        # Extract field components
        Ex_sel = E[0][..., mode_list]  # (Nx, Ny, 1, n)
        Ey_sel = E[1][..., mode_list]
        Hx_sel = H[0][..., mode_list]
        Hy_sel = H[1][..., mode_list]

        # Make the differential area elements
        Ex_Hy_dS = np.outer(dl_primal[0], dl_dual[1])
        Ey_Hx_dS = np.outer(dl_dual[0], dl_primal[1])

        # Initialize output matrix
        dtype = Ex_sel.dtype
        S = np.zeros((n_modes, n_modes), dtype=dtype)

        # Only compute upper triangle (including diagonal)
        for i in range(n_modes):
            # Vectorize over j >= i
            j_indices = np.arange(i, n_modes)

            # Extract mode i fields
            Ex_i = Ex_sel[..., i : i + 1]  # (Nx, Ny, 1, 1)
            Ey_i = Ey_sel[..., i : i + 1]
            Hy_i = Hy_sel[..., i : i + 1]
            Hx_i = Hx_sel[..., i : i + 1]

            # Extract mode j fields (vectorized for j >= i)
            Ex_j = Ex_sel[..., j_indices]  # (Nx, Ny, 1, n_j)
            Ey_j = Ey_sel[..., j_indices]
            Hy_j = Hy_sel[..., j_indices]
            Hx_j = Hx_sel[..., j_indices]

            # Compute term1: (Ex[i] * Hy[j] + Ex[j] * Hy[i]) * dS
            term1 = (Ex_i * Hy_j + Ex_j * Hy_i) * Ex_Hy_dS[..., np.newaxis]

            # Compute term2: (Ey[i] * Hx[j] + Ey[j] * Hx[i]) * dS
            term2 = (Ey_i * Hx_j + Ey_j * Hx_i) * Ey_Hx_dS[..., np.newaxis]

            # Sum over spatial dimensions to get S[i, j] for j >= i
            S[i, j_indices] = (1 / 4) * np.sum(term1 - term2, axis=(0, 1))

        # Fill lower triangle by symmetry
        S = S + S.T - np.diag(np.diag(S))

        return S

    @staticmethod
    def _normalize_modes(
        E: np.ndarray,
        H: np.ndarray,
        dl_primal: np.ndarray,
        dl_dual: np.ndarray,
        mode_indices: list[int],
        self_orthogonal_tol: float,
    ) -> tuple[np.ndarray, np.ndarray, set[int]]:
        """Normalize modes so that their self-dot equals 1.0.

        Returns:
            Tuple of (E, H, self_orthogonal_modes) where self_orthogonal_modes is a set of mode indices
            that are self-orthogonal and cannot be normalized.
        """
        self_orthogonal_modes = set()
        for mode_idx in mode_indices:
            self_dot = EigSolver._dot(E, H, dl_primal, dl_dual, mode_idx, mode_idx)
            self_dot_upper_bound = EigSolver._cauchy_schwarz_dot_bound(
                E, H, dl_primal, dl_dual, mode_idx
            )
            # Check if mode is self-orthogonal or otherwise not suitable for orthogonalization procedure
            if (
                np.abs(self_dot) < self_dot_upper_bound * self_orthogonal_tol
                or np.any(np.isnan(E[..., mode_idx]))
                or np.any(np.isnan(H[..., mode_idx]))
            ):
                self_orthogonal_modes.add(mode_idx)
                continue
            norm_factor = 1.0 / np.sqrt(self_dot)
            E[..., mode_idx] *= norm_factor
            H[..., mode_idx] *= norm_factor

        return E, H, self_orthogonal_modes

    @classmethod
    def _orthogonalize_degenerate_modes(
        cls,
        E: np.ndarray,
        H: np.ndarray,
        neff: np.ndarray,
        keff: np.ndarray,
        dl_primal: np.ndarray,
        dl_dual: np.ndarray,
        self_orthogonal_modes: set[int],
    ) -> tuple[np.ndarray, np.ndarray]:
        """Identify and orthogonalize degenerate modes, excluding self-orthogonal modes."""
        # Identify and orthogonalize candidate degenerate modes with a coarser tolerance
        degenerate_modes = cls._identify_degenerate_modes(
            neff + keff * 1j, TOL_DEGENERATE_CANDIDATE
        )

        # Remove self-orthogonal modes from degenerate mode sets since they cannot be orthogonalized
        degenerate_modes = [
            mode_set - self_orthogonal_modes
            for mode_set in degenerate_modes
            if len(mode_set - self_orthogonal_modes)
            > 1  # Only keep sets with at least 2 valid modes
        ]

        # Orthogonalize candidate degenerate modes that result in an overlap greater than
        # the required tolerance
        filtered_degenerate_modes = cls._identify_degenerate_modes_with_dot(
            degenerate_modes, E, H, dl_primal, dl_dual, TOL_NEEDS_ORTHOGONALIZATION
        )

        E, H = cls._make_orthogonal_basis_for_degenerate_modes(
            filtered_degenerate_modes, E, H, dl_primal, dl_dual
        )

        return E, H

    @staticmethod
    def _make_orthogonal_basis_for_degenerate_modes(
        degenerate_mode_sets: list[set[int]],
        E_vec: np.ndarray,
        H_vec: np.ndarray,
        dl_primal: np.ndarray,
        dl_dual: np.ndarray,
    ) -> tuple[np.ndarray, np.ndarray]:
        """Ensures that groups of degenerate modes are orthogonal, which is not guaranteed for eigenvectors with degenerate eigenvalues."""
        dtype = E_vec.dtype
        for degenerate_group in degenerate_mode_sets:
            # Convert set to list for indexing
            mode_indices = sorted(degenerate_group)

            # S is a symmetric overlap matrix of the degenerate modes.
            S = EigSolver._outer_dot(E_vec, H_vec, dl_primal, dl_dual, mode_indices)

            # 1. Diagonalize
            # eigenvalues (w) and right eigenvectors (v)
            eigvals, eigvecs = np.linalg.eig(S)
            # 2. Normalize Eigenvectors using UNCONJUGATED dot product
            # Standard np.linalg.eig normalizes so v.conj().T @ v = 1 (Hermitian)
            # We need v.T @ v = 1 (Complex Symmetric)
            for i in range(eigvecs.shape[1]):
                vec = eigvecs[:, i]
                # Calculate complex self-dot (no conjugation)
                self_dot = np.dot(vec, vec)

                # Avoid division by zero
                if np.isclose(self_dot, 0, atol=np.finfo(dtype).tiny):
                    raise ValueError(f"Eigenvector {i} is self-orthogonal. Cannot orthogonalize.")

                scale_factor = np.sqrt(self_dot)
                eigvecs[:, i] = vec / scale_factor
            # 3. Calculate Inverse Square Root of Eigenvalues
            lambda_inv_sqrt = np.diag(1.0 / np.sqrt(eigvals))
            # 4. Construct W
            # Since V is complex orthogonal (V.T @ V = I), V_inv = V.T
            # W = V * Lambda^(-1/2) * V.T
            W = eigvecs @ lambda_inv_sqrt @ eigvecs.T
            # S_new will be identity
            # S_new = W.T @ S @ W
            E_vec[..., mode_indices] = E_vec[..., mode_indices] @ W
            H_vec[..., mode_indices] = H_vec[..., mode_indices] @ W

        return E_vec, H_vec


def compute_modes(*args: Any, **kwargs: Any) -> tuple[Numpy, Numpy, str]:
    """A wrapper around ``EigSolver.compute_modes``, which is used in :class:`.ModeSolver`."""
    return EigSolver.compute_modes(*args, **kwargs)



"""Semiconductor doping definitions."""

from __future__ import annotations

from typing import Union

import numpy as np
import pydantic.v1 as pd
import xarray as xr

from tidy3d.components.autograd import TracedSize
from tidy3d.components.base import cached_property
from tidy3d.components.data.data_array import SpatialDataArray
from tidy3d.components.geometry.base import Box
from tidy3d.constants import MICROMETER, PERCMCUBE, inf
from tidy3d.exceptions import SetupError


class AbstractDopingBox(Box):
    """Derived class from Box to deal with dopings"""

    # Override size so that we can set default values
    size: TracedSize = pd.Field(
        (inf, inf, inf),
        title="Size",
        description="Size in x, y, and z directions.",
        units=MICROMETER,
    )

    def _get_indices_in_box(self, coords: dict, meshgrid: bool = True):
        """Returns locations inside box"""

        # work out whether x,y, and z are present
        dim_missing = len(list(coords.keys())) < 3
        if dim_missing:
            for var_name in "xyz":
                if var_name not in coords:
                    coords[var_name] = [0]

        if meshgrid:
            X, Y, Z = np.meshgrid(coords["x"], coords["y"], coords["z"], indexing="ij")
        else:
            X = coords["x"]
            Y = coords["y"]
            Z = coords["z"]

        new_bounds = [list(self.bounds[0]), list(self.bounds[1])]
        for d in range(3):
            if new_bounds[0][d] == new_bounds[1][d]:
                new_bounds[0][d] = -np.inf
                new_bounds[1][d] = np.inf

        # let's assume some of these coordinates may lay outside the box
        indices_in_box = np.logical_and(new_bounds[0][0] <= X, new_bounds[1][0] >= X)
        indices_in_box = np.logical_and(indices_in_box, new_bounds[0][1] <= Y)
        indices_in_box = np.logical_and(indices_in_box, new_bounds[1][1] >= Y)
        indices_in_box = np.logical_and(indices_in_box, new_bounds[0][2] <= Z)
        indices_in_box = np.logical_and(indices_in_box, new_bounds[1][2] >= Z)

        return indices_in_box, X, Y, Z

    def _post_init_validators(self) -> None:
        # check the doping box is 3D
        if len(self.zero_dims) > 0:
            raise SetupError(
                "The doping box must be 3D. If you want a 2D doping box, please set one of the dimensions to a large or infinite size."
            )


class ConstantDoping(AbstractDopingBox):
    """
    Sets constant doping :math:`N` in the specified box with a :py:attr:`~.Box.size` and :py:attr:`concentration`. For translationally invariant behavior in one dimension, the box must have infinite size in the
    homogenous (invariant) direction.

    Example
    -------
    >>> import tidy3d as td
    >>> box_coords = [
    ...     [-1, -1, -1],
    ...     [1, 1, 1]
    ... ]
    >>> constant_box1 = td.ConstantDoping(center=(0, 0, 0), size=(2, 2, 2), concentration=1e18)
    >>> constant_box2 = td.ConstantDoping.from_bounds(rmin=box_coords[0], rmax=box_coords[1], concentration=1e18)
    """

    concentration: pd.NonNegativeFloat = pd.Field(
        default=0,
        title="Doping concentration density.",
        description="Doping concentration density.",
        units=PERCMCUBE,
    )

    def _get_contrib(self, coords: dict, meshgrid: bool = True):
        """Returns the contribution to the doping a the locations specified in coords"""

        indices_in_box, X, _, _ = self._get_indices_in_box(coords=coords, meshgrid=meshgrid)

        contrib = np.zeros(X.shape)
        contrib[indices_in_box] = self.concentration

        return contrib.squeeze()


class GaussianDoping(AbstractDopingBox):
    """Sets a gaussian doping in the specified box. For translationally invariant behavior in one dimension, the box must have infinite size in the
    homogenous (invariant) direction.

    Notes
    -----
    The Gaussian doping concentration :math:`N` is defined in the following manner:

    - :math:`N=N_{\\text{max}}` at locations more than :math:`\\text{width}` um away from the sides of the box.
    - :math:`N=N_{\\text{ref}}` at location on the box sides.
    - a Gaussian variation between :math:`N_{\\text{max}}` and  :math:`N_{\\text{ref}}`  at locations less than :math:`\\text{width}`
    um away from the sides.

    By definition, all sides of the box will have concentration :math:`N_{\\text{ref}}` (except the side specified
    as source) and the center of the box (:math:`\\text{width}` away from the box sides) will have a concentration
    :math:`N_{\\text{max}}`.

    .. math::

        N = \\{N_{\\text{max}}\\} \\exp \\left[
        - \\ln \\left( \\frac{\\{N_{\\text{max}}\\}}{\\{N_{\\text{ref}}\\}} \\right)
        \\left( \\frac{(x|y|z) - \\{(x|y|z)_{\\text{box}}\\}}{\\text{width}} \\right)^2
        \\right]

    Example
    -------
    >>> import tidy3d as td
    >>> box_coords = [
    ...     [-1, -1, -1],
    ...     [1, 1, 1]
    ... ]
    >>> gaussian_box1 = td.GaussianDoping(
    ...     center=(0, 0, 0),
    ...     size=(2, 2, 2),
    ...     ref_con=1e15,
    ...     concentration=1e18,
    ...     width=0.1,
    ...     source="xmin"
    ... )
    >>> gaussian_box2 = td.GaussianDoping.from_bounds(
    ...     rmin=box_coords[0],
    ...     rmax=box_coords[1],
    ...     ref_con=1e15,
    ...     concentration=1e18,
    ...     width=0.1,
    ...     source="xmin"
    ... )
    """

    ref_con: pd.PositiveFloat = pd.Field(
        title="Reference concentration.",
        description="Reference concentration. This is the minimum concentration in the box "
        "and it is attained at the edges/faces of the box.",
        units=PERCMCUBE,
    )

    concentration: pd.PositiveFloat = pd.Field(
        title="Concentration",
        description="The concentration at the center of the box.",
        units=PERCMCUBE,
    )

    width: pd.PositiveFloat = pd.Field(
        title="Width of the gaussian.",
        description="Width of the gaussian. The concentration will transition from "
        "``concentration`` at the center of the box to ``ref_con`` at the edge/face "
        "of the box in a distance equal to ``width``. ",
        units=MICROMETER,
    )

    source: str = pd.Field(
        "xmin",
        title="Source face",
        description="Specifies the side of the box acting as the source, i.e., "
        "the face specified does not have a gaussian evolution normal to it, instead "
        "the concentration is constant from this face. Accepted values for ``source`` "
        "are [``xmin``, ``xmax``, ``ymin``, ``ymax``, ``zmin``, ``zmax``]",
    )

    @cached_property
    def sigma(self):
        """The sigma parameter of the pseudo-gaussian"""
        return np.sqrt(-self.width * self.width / 2 / np.log(self.ref_con / self.concentration))

    def _get_contrib(self, coords: dict, meshgrid: bool = True):
        """Returns the contribution to the doping a the locations specified in coords"""

        indices_in_box, X, Y, Z = self._get_indices_in_box(coords=coords, meshgrid=meshgrid)

        x_contrib = np.zeros(X.shape)
        x_contrib[indices_in_box] = 1.0
        if self.source != "xmin":
            x0 = self.bounds[0][0]
            indices = np.logical_and(x0 <= X, x0 + self.width >= X)
            indices = np.logical_and(indices, indices_in_box)
            x_contrib[indices] = np.exp(
                -(X[indices] - x0 - self.width)
                * (X[indices] - x0 - self.width)
                / 2
                / self.sigma
                / self.sigma
            )
        # higher x face
        if self.source != "xmax":
            x1 = self.bounds[1][0]
            indices = np.logical_and(x1 - self.width <= X, x1 >= X)
            indices = np.logical_and(indices, indices_in_box)
            x_contrib[indices] = np.exp(
                -(X[indices] - x1 + self.width)
                * (X[indices] - x1 + self.width)
                / 2
                / self.sigma
                / self.sigma
            )

        y_contrib = np.zeros(X.shape)
        y_contrib[indices_in_box] = 1.0
        if self.source != "ymin":
            y0 = self.bounds[0][1]
            indices = np.logical_and(y0 <= Y, y0 + self.width >= Y)
            indices = np.logical_and(indices, indices_in_box)
            y_contrib[indices] = np.exp(
                -(Y[indices] - y0 - self.width)
                * (Y[indices] - y0 - self.width)
                / 2
                / self.sigma
                / self.sigma
            )
        # higher y face
        if self.source != "ymax":
            y1 = self.bounds[1][1]
            indices = np.logical_and(y1 - self.width <= Y, y1 >= Y)
            indices = np.logical_and(indices, indices_in_box)
            y_contrib[indices] = np.exp(
                -(Y[indices] - y1 + self.width)
                * (Y[indices] - y1 + self.width)
                / 2
                / self.sigma
                / self.sigma
            )

        z_contrib = np.zeros(X.shape)
        z_contrib[indices_in_box] = 1.0
        if self.source != "zmin":
            z0 = self.bounds[0][2]
            indices = np.logical_and(z0 <= Z, z0 + self.width >= Z)
            indices = np.logical_and(indices, indices_in_box)
            z_contrib[indices] = np.exp(
                -(Z[indices] - z0 - self.width)
                * (Z[indices] - z0 - self.width)
                / 2
                / self.sigma
                / self.sigma
            )
        # higher z face
        if self.source != "zmax":
            z1 = self.bounds[1][2]
            indices = np.logical_and(z1 - self.width <= Z, z1 >= Z)
            indices = np.logical_and(indices, indices_in_box)
            z_contrib[indices] = np.exp(
                -(Z[indices] - z1 + self.width)
                * (Z[indices] - z1 + self.width)
                / 2
                / self.sigma
                / self.sigma
            )

        total_contrib = x_contrib * y_contrib * z_contrib * self.concentration

        return total_contrib.squeeze()


class CustomDoping(AbstractDopingBox):
    """Sets a custom doping in the specified box.

    Example
    -------
    >>> import tidy3d as td
    >>> import numpy as np
    >>> box_coords = [
    ...     [-1, -1, -1],
    ...     [1, 1, 1]
    ... ]
    >>> x = np.linspace(-1, 1, 5)
    >>> y = np.linspace(-1, 1, 5)
    >>> z = np.linspace(-1, 1, 5)
    >>> data = np.random.rand(5, 5, 5)*1e18
    >>> concentration = td.SpatialDataArray(
    ...     data=data,
    ...     coords={'x': x, 'y': y, 'z': z},
    ... )
    >>> custom_box1 = td.CustomDoping(
    ...     center=(0, 0, 0),
    ...     size=(2, 2, 2),
    ...     concentration=concentration
    ... )
    >>> custom_box2 = td.CustomDoping.from_bounds(
    ...     rmin=box_coords[0],
    ...     rmax=box_coords[1],
    ...     concentration=concentration
    ... )
    """

    concentration: SpatialDataArray = pd.Field(
        ...,
        title="Doping concentration data array.",
        description="Doping concentration data array.",
        units=PERCMCUBE,
    )

    def _get_contrib(self, coords: dict, meshgrid: bool = True):
        """Returns the contribution to the doping a the locations specified in coords"""

        indices_in_box, X, Y, Z = self._get_indices_in_box(coords=coords, meshgrid=meshgrid)

        contrib = np.zeros(X.shape)
        # interpolate
        if meshgrid:
            interp_result = self.concentration.interp(coords)
            contrib[indices_in_box] = interp_result.values[indices_in_box]
        else:
            # X, Y, Z are 1D arrays of coordinates
            # interp_result = self.concentration.interp(coords)
            # contrib = np.zeros(X.shape)
            # contrib[indices_in_box] = interp_result.values[indices_in_box]

            interp_coords = coords
            interp_da = {
                name: xr.DataArray(data, dims="new_dim") for name, data in interp_coords.items()
            }
            interp_res = self.concentration.interp(
                **interp_da, kwargs={"fill_value": 0, "bounds_error": False}
            )
            contrib[indices_in_box] = interp_res.values[indices_in_box]

        return contrib.squeeze()


DopingBoxType = Union[ConstantDoping, GaussianDoping, CustomDoping]



"""Defines information about a task"""

from __future__ import annotations

from abc import ABC
from datetime import datetime
from enum import Enum
from typing import Optional

import pydantic.v1 as pydantic


class TaskBase(pydantic.BaseModel, ABC):
    """Base configuration for all task objects."""

    class Config:
        """Configuration for TaskBase"""

        arbitrary_types_allowed = True
        """Allow arbitrary types to be used within the model."""


class ChargeType(str, Enum):
    """The payment method of the task."""

    FREE = "free"
    """No payment required."""

    PAID = "paid"
    """Payment required."""


class TaskBlockInfo(TaskBase):
    """Information about the task's block status.

    Notes
    -----
        This includes details about how the task can be blocked by various features
        such as user limits and insufficient balance.
    """

    chargeType: ChargeType = None
    """The type of charge applicable to the task (free or paid)."""

    maxFreeCount: int = None
    """The maximum number of free tasks allowed."""

    maxGridPoints: int = None
    """The maximum number of grid points permitted."""

    maxTimeSteps: int = None
    """The maximum number of time steps allowed."""


class TaskInfo(TaskBase):
    """General information about a task."""

    taskId: str
    """Unique identifier for the task."""

    taskName: str = None
    """Name of the task."""

    nodeSize: int = None
    """Size of the node allocated for the task."""

    completedAt: Optional[datetime] = None
    """Timestamp when the task was completed."""

    status: str = None
    """Current status of the task."""

    realCost: float = None
    """Actual cost incurred by the task."""

    timeSteps: int = None
    """Number of time steps involved in the task."""

    solverVersion: str = None
    """Version of the solver used for the task."""

    createAt: Optional[datetime] = None
    """Timestamp when the task was created."""

    estCostMin: float = None
    """Estimated minimum cost for the task."""

    estCostMax: float = None
    """Estimated maximum cost for the task."""

    realFlexUnit: float = None
    """Actual flexible units used by the task."""

    oriRealFlexUnit: float = None
    """Original real flexible units."""

    estFlexUnit: float = None
    """Estimated flexible units for the task."""

    estFlexCreditTimeStepping: float = None
    """Estimated flexible credits for time stepping."""

    estFlexCreditPostProcess: float = None
    """Estimated flexible credits for post-processing."""

    estFlexCreditMode: float = None
    """Estimated flexible credits based on the mode."""

    s3Storage: float = None
    """Amount of S3 storage used by the task."""

    startSolverTime: Optional[datetime] = None
    """Timestamp when the solver started."""

    finishSolverTime: Optional[datetime] = None
    """Timestamp when the solver finished."""

    totalSolverTime: int = None
    """Total time taken by the solver."""

    callbackUrl: str = None
    """Callback URL for task notifications."""

    taskType: str = None
    """Type of the task."""

    metadataStatus: str = None
    """Status of the metadata for the task."""

    taskBlockInfo: TaskBlockInfo = None
    """Blocking information for the task."""

    version: str = None
    """Version of the task."""


class RunInfo(TaskBase):
    """Information about the run of a task."""

    perc_done: pydantic.confloat(ge=0.0, le=100.0)
    """Percentage of the task that is completed (0 to 100)."""

    field_decay: pydantic.confloat(ge=0.0, le=1.0)
    """Field decay from the maximum value (0 to 1)."""

    def display(self) -> None:
        """Print some info about the task's progress."""
        print(f" - {self.perc_done:.2f} (%) done")
        print(f" - {self.field_decay:.2e} field decay from max")


# ---------------------- Batch (Modeler) detail schema ---------------------- #


class BatchTaskBlockInfo(TaskBlockInfo):
    """
    Extends `TaskBlockInfo` with specific details for batch task blocking.

    Attributes:
        accountLimit: A usage or cost limit imposed by the user's account.
        taskBlockMsg: A human-readable message describing the reason for the block.
        taskBlockType: The specific type of block (e.g., 'balance', 'limit').
        blockStatus: The current blocking status for the batch.
        taskStatus: The status of the task when it was blocked.
    """

    accountLimit: float = None
    taskBlockMsg: str = None
    taskBlockType: str = None
    blockStatus: str = None
    taskStatus: str = None


class BatchMember(TaskBase):
    """
    Represents a single task within a larger batch operation.

    Attributes:
        refId: A reference identifier for the member task.
        folderId: The identifier of the folder containing the task.
        sweepId: The identifier for the parameter sweep, if applicable.
        taskId: The unique identifier of the task.
        linkedTaskId: The identifier of a task linked to this one.
        groupId: The identifier of the group this task belongs to.
        taskName: The name of the individual task.
        status: The current status of this specific task.
        sweepData: Data associated with a parameter sweep.
        validateInfo: Information related to the task's validation.
        replaceData: Data used for replacements or modifications.
        protocolVersion: The version of the protocol used.
        variable: The variable parameter for this task in a sweep.
        createdAt: The timestamp when the member task was created.
        updatedAt: The timestamp when the member task was last updated.
        denormalizeStatus: The status of the data denormalization process.
        summary: A dictionary containing summary information for the task.
    """

    refId: str = None
    folderId: str = None
    sweepId: str = None
    taskId: str = None
    linkedTaskId: str = None
    groupId: str = None
    taskName: str = None
    status: str = None
    sweepData: str = None
    validateInfo: str = None
    replaceData: str = None
    protocolVersion: str = None
    variable: str = None
    createdAt: Optional[datetime] = None
    updatedAt: Optional[datetime] = None
    denormalizeStatus: str = None
    summary: dict = None


class BatchDetail(TaskBase):
    """Provides a detailed, top-level view of a batch of tasks.

    Notes
    -----
        This model serves as the main payload for retrieving comprehensive
        information about a batch operation.

    Attributes
    ----------
    refId
        A reference identifier for the entire batch.
    optimizationId
        Identifier for the optimization process, if any.
    groupId
        Identifier for the group the batch belongs to.
    name
        The user-defined name of the batch.
    status
        The current status of the batch.
    totalTask
        The total number of tasks in the batch.
    preprocessSuccess
        The count of tasks that completed preprocessing.
    postprocessStatus
        The status of the batch's postprocessing stage.
    validateSuccess
        The count of tasks that passed validation.
    runSuccess
        The count of tasks that ran successfully.
    postprocessSuccess
        The count of tasks that completed postprocessing.
    taskBlockInfo
        Information on what might be blocking the batch.
    estFlexUnit
        The estimated total flexible compute units for the batch.
    totalSeconds
        The total time in seconds the batch has taken.
    totalCheckMillis
        Total time in milliseconds spent on checks.
    message
        A general message providing information about the batch status.
    tasks
        A list of `BatchMember` objects, one for each task in the batch.
    taskType
        The type of tasks contained in the batch.
    """

    refId: str = None
    optimizationId: str = None
    groupId: str = None
    name: str = None
    status: str = None
    totalTask: int = 0
    preprocessSuccess: int = 0
    postprocessStatus: str = None
    validateSuccess: int = 0
    runSuccess: int = 0
    postprocessSuccess: int = 0
    taskBlockInfo: BatchTaskBlockInfo = None
    estFlexUnit: float = None
    realFlexUnit: float = None
    totalSeconds: int = None
    totalCheckMillis: int = None
    message: str = None
    tasks: list[BatchMember] = []
    validateErrors: dict = None
    taskType: str = None
    version: str = None


class AsyncJobDetail(TaskBase):
    """Provides a detailed view of an asynchronous job and its sub-tasks.

    Notes
    -----
        This model represents a long-running operation. The 'result' attribute holds
        the output of a completed job, which for orchestration jobs, is often a
        JSON string mapping sub-task names to their unique IDs.

    Attributes
    ----------
    asyncId
        The unique identifier for the asynchronous job.
    status
        The current overall status of the job (e.g., 'RUNNING', 'COMPLETED').
    progress
        The completion percentage of the job (from 0.0 to 100.0).
    createdAt
        The timestamp when the job was created.
    completedAt
        The timestamp when the job finished (successfully or not).
    tasks
        A dictionary mapping logical task keys to their unique task IDs.
        This is often populated by parsing the 'result' of an orchestration task.
    result
        The raw string output of the completed job. If the job spawns other
        tasks, this is expected to be a JSON string detailing those tasks.
    taskBlockInfo
        Information on any dependencies blocking the job from running.
    message
        A human-readable message about the job's status.
    """

    asyncId: str
    status: str
    progress: Optional[float] = None
    createdAt: Optional[datetime] = None
    completedAt: Optional[datetime] = None
    tasks: Optional[dict[str, str]] = None
    result: Optional[str] = None
    taskBlockInfo: Optional[TaskBlockInfo] = None
    message: Optional[str] = None


AsyncJobDetail.update_forward_refs()



from __future__ import annotations

from typing import Union

from .ac import SSACVoltageSource
from .dc import DCCurrentSource, DCVoltageSource, GroundVoltage

VoltageSourceType = Union[DCVoltageSource, SSACVoltageSource, GroundVoltage]
CurrentSourceType = Union[DCCurrentSource]
